/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <limits.h>
#include <complib/cl_qhashmap.h>
#include <complib/cl_hashmap.h>

/******************************************************************************
 IMPLEMENTATION OF QHASHMAP
******************************************************************************/

/*
 * Convenience macro for readability.
 */
#define foreach_bucket(P_HASH,P_BUCKET)					\
        for ((P_BUCKET) = (P_HASH)->buckets;				\
	     (P_BUCKET) < (P_HASH)->buckets + (P_HASH)->num_buckets;	\
	     (P_BUCKET)++)

#define foreach_bucket_item(P_BUCKET, P_ITEM)				\
	for (P_ITEM  = P_BUCKET->list;				\
	     P_ITEM != NULL;						\
	     P_ITEM  = P_ITEM->next)

#define	HASHMAP_INIT_COUNT		32

/* ---------------------------------------------------------------------------------------------- */

/*
 * Simple, fast hash functions.
 */
static inline uint64_t _cl_f_hash_uint64(uintptr_t key, size_t length)
{
	return (uint64_t)key;
}

static inline int _cl_f_cmp_uint64(uintptr_t lhs, uintptr_t rhs, size_t length)
{
	return lhs == rhs;
}

static inline uint64_t _cl_f_hash_memory(uintptr_t key, size_t length)
{
#define	P	877ull			// Prime for the rolling hash
#define	M	0x100000001B3ull	// Modulus for rolling hash
#define	S	32			// size of the prime array

	static uint64_t		p[S];
	int			i;
	int			j;
	uint64_t		x;
	char			*cp;

	/*
	 * Initialize the rolling hash array.
	 */
	if (p[0] == 0) {
		p[0] = P;
		for (i = 1; i < S; i++) {
			p[i] = (P*p[i-1]) % M;
		}
	}

	i = 0;
	x = 0;
	cp = (char *)key;
	for (j = 0; j < length; j++) {
		x += cp[j] * p[i];
		i = (i+1) & (S-1);
	}

	return (x << 37) + (x) + (x >> 15);
}

static inline int _cl_f_cmp_memory(uintptr_t lhs, uintptr_t rhs, size_t length)
{
	return !memcmp((void *)lhs, (void *)rhs, length);
}

static inline uint64_t _cl_f_hash_string(uintptr_t key, size_t length)
{
	return _cl_f_hash_memory(key, strlen((char *)key));
}

static inline int _cl_f_cmp_string(uintptr_t lhs, uintptr_t rhs, size_t length)
{
	return !strcmp((char *)lhs, (char *)rhs);
}
/* ---------------------------------------------------------------------------------------------- */

/* QHASHMAP Internal functions */

/*
 * When the table gets "too full", it needs to be resized.
 */
#define	HASHMAP_RESIZE_FACTOR		 5
#define	HASHMAP_MAX_BUCKET_SIZE		 9

static inline boolean_t _cl_qhashmap_check(IN const cl_qhashmap_t * const p_hash)
{
	return ((p_hash != NULL) && (p_hash->num_buckets > 0));
}

static inline boolean_t _is_resize_needed(IN cl_qhashmap_t *p_hash, cl_qhashmap_bucket_t *p_bucket)
{

	if (p_bucket->count > HASHMAP_MAX_BUCKET_SIZE) {
		return TRUE;
	} else if ((p_hash->count/p_hash->num_buckets) > HASHMAP_RESIZE_FACTOR) {
		return TRUE;
	} else {
		return FALSE;
	}
}

static size_t _cl_qhashmap_new_count(size_t current_count)
{
	static long primes[] = {
		(1ull <<  5) +  5, (1ull <<  6) +  3, (1ull <<  7) +  3, (1ull <<  8) +  1,
		(1ull <<  9) +  9, (1ull << 10) +  7, (1ull << 11) +  5, (1ull << 12) +  3,
		(1ull << 13) + 17, (1ull << 14) + 27, (1ull << 15) +  3, (1ull << 16) +  1,
		(1ull << 17) + 29, (1ull << 18) +  3, (1ull << 19) + 21, (1ull << 20) +  7,
		(1ull << 21) + 17, (1ull << 22) + 15, (1ull << 23) +  9, (1ull << 24) + 43,
		(1ull << 25) + 35, (1ull << 26) + 15, (1ull << 27) + 29, (1ull << 28) +  3,
		(1ull << 29) + 11, (1ull << 30) +  3, (1ull << 31) + 11, (1ull << 32) +  1,
	};

        int		i;

	for (i = 0; i < sizeof(primes) / sizeof(primes[0]); i++) {
		if (primes[i] > current_count) {
			return primes[i];
		}
	}

	return 0;
}


static inline cl_qhashmap_bucket_t *_cl_qhashmap_key_to_bucket(IN cl_qhashmap_t const *p_hash,
							     IN uint64_t key)
{
	uint64_t		index;

	if (p_hash->num_buckets == 0) {
		return NULL;
	}

	index = p_hash->f_hash(key, p_hash->key_length) % p_hash->num_buckets;
	return &p_hash->buckets[index];
}

static inline void _cl_qhashmap_insert_item(IN cl_qhashmap_t *p_hash,
					   IN cl_qhashmap_item_t *p_item)
{
	cl_qhashmap_item_t	*p_next;
	cl_qhashmap_bucket_t	*p_bucket;

	/*
	 * Update table and bucket.  Place the item at the head of the bucket list.
	 */
	if ((p_bucket = _cl_qhashmap_key_to_bucket(p_hash, p_item->key)) == NULL) {
		return;
	}

	p_next = p_bucket->list;

	p_item->b_index = p_bucket->index;
	p_item->prev = NULL;
	p_item->next = p_next;

	if (p_next != NULL) {
		p_next->prev = p_item;
	}

	p_bucket->list = p_item;
	p_bucket->count++;
	p_hash->count++;
}

static inline cl_status_t _cl_qhashmap_remove_item(IN cl_qhashmap_t * const p_hash,
						  IN cl_qhashmap_item_t *p_item)
{
	cl_qhashmap_item_t	*p_find;
	cl_qhashmap_item_t	*p_prev;
	cl_qhashmap_item_t	*p_next;
	cl_qhashmap_bucket_t	*p_bucket;

	if ((p_hash == NULL) || (p_item == NULL)) {
		return CL_INVALID_PARAMETER;
	}

	/*
	 * Get the bucket.  Check that the item is in the bucket.
	 */
	if (p_item->b_index >= p_hash->num_buckets) {
		return CL_ERROR;
	}

	p_bucket = &p_hash->buckets[p_item->b_index];
	for (p_find = p_bucket->list; p_find != NULL; p_find = p_find->next) {
		if (p_find == p_item) {
			break;
		}
	}

	if (p_find == NULL) {
		return CL_ERROR;
	}

	/*
	 * Relink the bucket.
	 */
	p_prev = p_item->prev;
	p_next = p_item->next;

	p_bucket->count--;
	p_hash->count--;

	if (p_prev == NULL) {
		p_bucket->list = p_next;
	} else {
		p_prev->next = p_next;
	}

	if (p_next != NULL) {
		p_next->prev = p_prev;
	}

	p_item->prev = NULL;
	p_item->next = NULL;
	p_item->b_index = -1;
	return CL_SUCCESS;
}

static cl_status_t _cl_qhashmap_init(IN cl_qhashmap_t * const p_hash,
				     IN size_t bucket_count,
				     IN cl_qhashmap_type_t type,
				     IN size_t key_length)
{
	size_t			num_buckets;
	uint64_t		(*hash_function)(uintptr_t, size_t);
	int			(*cmp_function)(uintptr_t, uintptr_t, size_t);
	cl_qhashmap_bucket_t	*p_bucket;
	cl_qhashmap_bucket_t	*p_buckets;

	/*
	 * Parameter checking.
	 */
	if (p_hash == NULL) {
		return CL_INVALID_PARAMETER;
	}

	if (type == cl_qhashmap_type_uint64) {
		hash_function = _cl_f_hash_uint64;
		cmp_function = _cl_f_cmp_uint64;
	} else if (type == cl_qhashmap_type_memory) {
		hash_function = _cl_f_hash_memory;
		cmp_function = _cl_f_cmp_memory;
	} else if (type == cl_qhashmap_type_string) {
		hash_function = _cl_f_hash_string;
		cmp_function = _cl_f_cmp_string;
	} else {
		return CL_INVALID_PARAMETER;
	}

	/*
	 * Get the number of buckets for hashing.
	 */
	if ((num_buckets = _cl_qhashmap_new_count(bucket_count)) == 0) {
		return CL_INVALID_PARAMETER;
	}

	memset(p_hash, 0, sizeof(cl_qhashmap_t));
	p_buckets = calloc(num_buckets, sizeof(cl_qhashmap_bucket_t));
	if (p_buckets == NULL) {
		return CL_INSUFFICIENT_MEMORY;
	}

	/*
	 * initialize the buckets.
	 */
	p_hash->type = type;
	p_hash->count = 0;
	p_hash->f_hash = hash_function;
	p_hash->f_cmp = cmp_function;
	p_hash->num_buckets = num_buckets;
	p_hash->buckets = p_buckets;
	p_hash->key_length = key_length;

	foreach_bucket(p_hash, p_bucket) {
		p_bucket->list = NULL;
		p_bucket->index = p_bucket - p_hash->buckets;
		p_bucket->count = 0;
	}

	return CL_SUCCESS;
}

static void _cl_qhashmap_resize(IN cl_qhashmap_t *p_hash)
{
	cl_qhashmap_t		old_hash;
	cl_qhashmap_item_t	*p_item;
	cl_qhashmap_bucket_t	*p_bucket;

	/*
	 * Save the current hashmap.  We will use this to free the resources later.
	 */
	memcpy((void *)&old_hash, (void *)p_hash, sizeof(cl_qhashmap_t));

	/*
	 * Initialize the new hashmap for moving the old items into. If for some reason
	 * we can't initialize a new map for resizing, just keep the old one.
	 */
	if (_cl_qhashmap_init(p_hash, old_hash.num_buckets, old_hash.type, old_hash.key_length) != CL_SUCCESS) {
		return;
	}

	/*
	 * Move the items to the new buckets.
	 */
	foreach_bucket(&old_hash, p_bucket) {
		while (p_bucket->list != NULL) {
			p_item = p_bucket->list;
			p_bucket->list = p_item->next;
			_cl_qhashmap_insert_item(p_hash, p_item);
		}
	}

	/*
	 * Clean up the old hashmap.  Everything should have been rehashed by this point.
	 * The buckets and the items in the buckets have already been transferred to the
	 * resized hashmap.  All that remains is to free the alloc-ed memory.
	 */
	cl_qhashmap_destroy(&old_hash);
	return;
}

static cl_qhashmap_item_t *_cl_qhashmap_find_key(IN const cl_qhashmap_t * const p_hash,
					 IN uintptr_t key)
{
	cl_qhashmap_item_t	*p_item;
	cl_qhashmap_bucket_t	*p_bucket;

	if ((p_bucket = _cl_qhashmap_key_to_bucket(p_hash, key)) == NULL) {
		return NULL;
	}

	foreach_bucket_item(p_bucket, p_item) {
		if (p_hash->f_cmp(key, p_item->key, p_hash->key_length)) {
			return p_item;
		}
	}

        return NULL;
}

/* ---------------------------------------------------------------------------------------------- */

/* QHASHMAP API functions */

cl_status_t cl_qhashmap_init_uint64(IN cl_qhashmap_t * const p_hash)
{
	CL_ASSERT(p_hash != NULL);

	return _cl_qhashmap_init(p_hash, HASHMAP_INIT_COUNT, cl_qhashmap_type_uint64, 0);
}

cl_status_t cl_qhashmap_init_memory(IN cl_qhashmap_t * const p_hash, IN size_t length)
{
	CL_ASSERT(p_hash != NULL);
	CL_ASSERT(length > 0);

	return _cl_qhashmap_init(p_hash, HASHMAP_INIT_COUNT, cl_qhashmap_type_memory, length);
}

cl_status_t cl_qhashmap_init_string(IN cl_qhashmap_t * const p_hash)
{
	CL_ASSERT(p_hash != NULL);

	return _cl_qhashmap_init(p_hash, HASHMAP_INIT_COUNT, cl_qhashmap_type_string, 0);
}

cl_status_t cl_qhashmap_destroy(IN cl_qhashmap_t * const p_hash)
{
	CL_ASSERT(_cl_qhashmap_check(p_hash));

	/*
	 * Free items which are being held:
	 *	1) hashed cl_qhashmap_item_t in buckets.
	 *	2) buckets themselves.
	 */
	cl_qhashmap_remove_all(p_hash);
	free(p_hash->buckets);

	memset(p_hash, 0, sizeof(cl_qhashmap_t));
	return CL_SUCCESS;
}

cl_qhashmap_item_t *cl_qhashmap_insert(IN cl_qhashmap_t *p_hash,
				       IN cl_qhashmap_item_t *p_item)
{
	cl_qhashmap_bucket_t	*p_bucket;
	cl_qhashmap_item_t	*p_find;

	CL_ASSERT(_cl_qhashmap_check(p_hash));
	CL_ASSERT(p_item != NULL);

	/*
	 * See if the key already exists in the map.
	 */
	if ((p_find = cl_qhashmap_find(p_hash, p_item->key)) != NULL) {
		return p_find;
	}

	_cl_qhashmap_insert_item(p_hash, p_item);

	/*
	 * Check if this insertion triggers a resize.
	 */
	p_bucket = &p_hash->buckets[p_item->b_index];
	if (_is_resize_needed(p_hash, p_bucket)) {
		_cl_qhashmap_resize(p_hash);
	}

	return p_item;
}

cl_status_t cl_qhashmap_remove(IN cl_qhashmap_t * const p_hash,
			       IN cl_qhashmap_item_t *p_item)
{
	CL_ASSERT(_cl_qhashmap_check(p_hash));
	CL_ASSERT(p_item != NULL);

	return _cl_qhashmap_remove_item(p_hash, p_item);
}

cl_status_t cl_qhashmap_remove_all(IN cl_qhashmap_t * const p_hash)
{
	cl_qhashmap_bucket_t	*p_bucket;

	CL_ASSERT(_cl_qhashmap_check(p_hash));

	/*
	 * Clean up each bucket individually.  Since each bucket is a list, we
	 * just call the qlist remove all function to do our work for us.
	 */
	foreach_bucket(p_hash, p_bucket) {
		while (p_bucket->list != NULL) {
			_cl_qhashmap_remove_item(p_hash, p_bucket->list);
		}
	}

	return CL_SUCCESS;
}

void cl_qhashmap_apply_func(IN const cl_qhashmap_t * const p_hash,
                            IN cl_qhashmap_apply_t p_func,
                            IN void * const context)
{
	cl_qhashmap_item_t	*p_item;
	cl_qhashmap_bucket_t	*p_bucket;

	CL_ASSERT(_cl_qhashmap_check(p_hash));
	CL_ASSERT(p_func != NULL);

	/*
	 * Apply the function to each item as we loop over the buckets.
	 */
	foreach_bucket(p_hash, p_bucket) {
		foreach_bucket_item(p_bucket, p_item) {
			p_func(p_item, context);
		}
	}

	return;
}

cl_qhashmap_item_t *cl_qhashmap_find(IN const cl_qhashmap_t * const p_hash,
				     IN uintptr_t key)
{
	CL_ASSERT(_cl_qhashmap_check(p_hash));

	return _cl_qhashmap_find_key(p_hash, key);
}

/* ---------------------------------------------------------------------------------------------- */

/* HASHMAP Internal functions */

#define	HM_MAX_FREE_COUNT		256

#define HM_ENQUEUE(P_HASH, P_ITEM)						\
	P_ITEM->prev = NULL;							\
	P_ITEM->next = P_HASH->free;						\
	P_HASH->free = P_ITEM;							\
	P_HASH->free_count++

#define HM_DEQUEUE(P_HASH, P_ITEM)						\
	P_ITEM = P_HASH->free;							\
	P_HASH->free = P_ITEM->next;						\
	P_HASH->free_count--


static inline boolean_t _cl_hashmap_check(IN const cl_hashmap_t * const p_hash)
{
	return (p_hash != NULL);
}

/*
 * The free list has items which have been previously inserted and removed.  We keep a number
 * around so we don't have to do malloc every time.
 */
static inline cl_qhashmap_item_t *_cl_hashmap_alloc(IN cl_hashmap_t * const p_hash)
{
	cl_qhashmap_item_t	*p_item;

	if (p_hash->free != NULL) {
		HM_DEQUEUE(p_hash, p_item);
		return p_item;
	}

	return (cl_qhashmap_item_t *)calloc(1, sizeof(*p_item));
}

static inline void _cl_hashmap_free(IN cl_hashmap_t * const p_hash,
				    IN cl_qhashmap_item_t *p_item)
{
	if (p_hash->free_count < HM_MAX_FREE_COUNT) {
		HM_ENQUEUE(p_hash, p_item);
	} else {
		free(p_item);
	}
}

/* ---------------------------------------------------------------------------------------------- */

/* HASHMAP API functions */


cl_status_t cl_hashmap_init_uint64(IN cl_hashmap_t * const p_hash)
{
	CL_ASSERT(_cl_hashmap_check(p_hash));

	memset((void *)p_hash, 0, sizeof(cl_hashmap_t));
	return cl_qhashmap_init_uint64(&p_hash->hashmap);
}

cl_status_t cl_hashmap_init_memory(IN cl_hashmap_t * const p_hash, IN size_t length)
{
	CL_ASSERT(_cl_hashmap_check(p_hash));

	memset((void *)p_hash, 0, sizeof(cl_hashmap_t));
	return cl_qhashmap_init_memory(&p_hash->hashmap, length);
}

cl_status_t cl_hashmap_init_string(IN cl_hashmap_t * const p_hash)
{
	CL_ASSERT(_cl_hashmap_check(p_hash));

	memset((void *)p_hash, 0, sizeof(cl_hashmap_t));
	return cl_qhashmap_init_string(&p_hash->hashmap);
}

void cl_hashmap_destroy(IN cl_hashmap_t * const p_hash)
{
	cl_qhashmap_item_t	*p_item;

	CL_ASSERT(_cl_hashmap_check(p_hash));

	cl_hashmap_remove_all(p_hash);
	while (p_hash->free_count > 0) {
		HM_DEQUEUE(p_hash, p_item);
		free(p_item);
	}

	cl_qhashmap_destroy(&p_hash->hashmap);
	memset(p_hash, 0, sizeof(cl_hashmap_t));
}

cl_status_t cl_hashmap_insert(IN cl_hashmap_t *p_hash,
			      IN uint64_t key,
			      IN uintptr_t value)
{
	cl_qhashmap_item_t	*p_item;

	CL_ASSERT(_cl_hashmap_check(p_hash));

	/*
	 * See if the key already exists in the map.
	 */
	if (cl_qhashmap_find(&p_hash->hashmap, key) != NULL) {
		return CL_DUPLICATE;
	}

	/*
	 * Create the item to enqueue.
	 */
	if ((p_item = _cl_hashmap_alloc(p_hash)) == NULL) {
		return CL_INSUFFICIENT_MEMORY;
	}

	p_item->key = key;
	p_item->value = value;
	cl_qhashmap_insert(&p_hash->hashmap, p_item);

	return CL_SUCCESS;
}

cl_status_t cl_hashmap_remove(IN cl_hashmap_t * const p_hash,
			      IN uint64_t key)
{
	cl_status_t		status;
	cl_qhashmap_item_t	*p_item;

	CL_ASSERT(_cl_hashmap_check(p_hash));

	/*
	 * See if the key exists in the map.
	 */
	if ((p_item = cl_qhashmap_find(&p_hash->hashmap, key)) == NULL) {
		return CL_ERROR;
	}

	/*
	 * Remove the item and enqueue it.
	 */
	status = cl_qhashmap_remove(&p_hash->hashmap, p_item);
	if (status == CL_SUCCESS) {
		_cl_hashmap_free(p_hash, p_item);
	}

	return status;
}

void cl_hashmap_remove_all(IN cl_hashmap_t * const p_hash)
{
	cl_qhashmap_item_t	*p_item;
	cl_qhashmap_bucket_t	*p_bucket;

	CL_ASSERT(_cl_hashmap_check(p_hash));

	/*
	 * We can't call cl_qhashmap_remove_all because we would
	 * lose track of the items which are discarded.
	 */
	foreach_bucket(&p_hash->hashmap, p_bucket) {
		while ((p_item = p_bucket->list) != NULL) {
			cl_qhashmap_remove(&p_hash->hashmap, p_item);
			_cl_hashmap_free(p_hash, p_item);
		}
	}
}

void cl_hashmap_apply_func(IN cl_hashmap_t * const p_hash,
                           IN cl_qhashmap_apply_t p_func,
                           IN void * const context)
{
	CL_ASSERT(_cl_hashmap_check(p_hash));
	CL_ASSERT(p_func != NULL);

	cl_qhashmap_apply_func(&p_hash->hashmap, p_func, context);
	return;
}

uintptr_t cl_hashmap_find(IN cl_hashmap_t * const p_hash,
			  IN uintptr_t key)
{
	cl_qhashmap_item_t	*p_item;

	CL_ASSERT(_cl_hashmap_check(p_hash));

	if ((p_item = cl_qhashmap_find(&p_hash->hashmap, key)) != NULL) {
		return p_item->value;
	}

	return 0;
}
