%{
/*
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_CC_ALGO_PARSER_Y_Y
#include <opensm/osm_opensm.h>

#define YYSTYPE char * 
#define PPCC_ALGO_MAX_LINE_LEN             4096
#define PPCC_ALGO_ID_BITS_NUM              16
#define PPCC_ALGO_MAJOR_VERSION_BITS_NUM   8
#define PPCC_ALGO_MINOR_VERSION_BITS_NUM   8

typedef struct ppcc_algo_tmp_parser_struct_ {
	cl_list_t  str_list;
	cl_list_t  num_list;
	osm_ppcc_dictionary_entry_t dictionary_array[PPCC_ALGO_MAX_PARAMS];
	int current_index_in_dictionary;
	ib_ppcc_algo_info_t version;
	osm_ppcc_dictionary_entry_t entry;
	char file_name[PPCC_ALGO_MAX_FILE_NAME];
} ppcc_algo_tmp_parser_struct_t;

typedef enum ca_algo_conf_id {
	CA_ALGO_CONFIG_NAME,
	CA_ALGO_CONFIG_VERSION,
	CA_ALGO_CONFIG_PARAMS_NAME_LIST,
	CA_ALGO_CONFIG_PARAMS_MIN_VALS,
	CA_ALGO_CONFIG_PARAMS_MAX_VALS,
	CA_ALGO_CONFIG_LAST,
} ca_algo_conf_id_t;

typedef enum algo_conf_kind {
	CONF_CA_ALGO,
} algo_conf_kind_t;

extern FILE * __cc_algo_parser_in;
extern char *trim_whitespace(char *str);
extern void __cc_algo_parser_restart(FILE *input_file);

static struct osm_opensm *p_osm;
boolean_t ca_algo_conf_tags_parsed[CA_ALGO_CONFIG_LAST];

char *__cc_algo_parser_yylval = NULL;

ppcc_algo_tmp_parser_struct_t ppcc_algo_tmp_parser_struct;

int cc_algo_column_num;
int cc_algo_line_num;

osm_cc_ca_algo_conf_t *p_current_ca_algo_conf = NULL;


static void parser_tmp_struct_init(char *file_name)
{
	ppcc_algo_tmp_parser_struct.entry.str[0] = '\0';
	ppcc_algo_tmp_parser_struct.current_index_in_dictionary = 0;
	memcpy(ppcc_algo_tmp_parser_struct.file_name, file_name,
	       strlen(file_name) + 1);
	cl_list_construct(&ppcc_algo_tmp_parser_struct.str_list);
	cl_list_init(&ppcc_algo_tmp_parser_struct.str_list, 10);
	cl_list_construct(&ppcc_algo_tmp_parser_struct.num_list);
	cl_list_init(&ppcc_algo_tmp_parser_struct.num_list, 10);
}

/*
 * Do NOT free objects from the temp struct.
 * Either they are inserted into the parse tree data
 * structure, or they are already freed when copying
 * their values to the parse tree data structure.
 */
static void parser_tmp_struct_reset()
{
	ppcc_algo_tmp_parser_struct.entry.str[0] = '\0';
	ppcc_algo_tmp_parser_struct.current_index_in_dictionary = 0;
	cl_list_remove_all(&ppcc_algo_tmp_parser_struct.str_list);
	cl_list_remove_all(&ppcc_algo_tmp_parser_struct.num_list);
}

static void parser_tmp_struct_destroy()
{
	parser_tmp_struct_reset();
	cl_list_destroy(&ppcc_algo_tmp_parser_struct.str_list);
	cl_list_destroy(&ppcc_algo_tmp_parser_struct.num_list);
}

static void clear_val(char **c)
{
	free(*c);
	*c = NULL;
}

static void yyerror(const char *format, ...)
{
    char s[256];
    va_list pvar;

    OSM_LOG_ENTER(&p_osm->log);

    va_start(pvar, format);
    vsnprintf(s, sizeof(s), format, pvar);
    va_end(pvar);

    OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9E09: "
            "Syntax error (line %d:%d): %s\n",
            cc_algo_line_num, cc_algo_column_num, s);
    fprintf(stderr, "Error in CC Algo Policy File (line %d:%d): %s.\n",
            cc_algo_line_num, cc_algo_column_num, s);
    OSM_LOG_EXIT(&p_osm->log);
}

const char *get_ca_algo_param_name_by_id(int param_id)
{
	switch (param_id) {
	case CA_ALGO_CONFIG_NAME:
		return "ca_algo_name";
	case CA_ALGO_CONFIG_VERSION:
		return "ca_algo_version";
	case CA_ALGO_CONFIG_PARAMS_NAME_LIST:
		return "ca_algo_config_params_name_list";
	case CA_ALGO_CONFIG_PARAMS_MIN_VALS:
		return "ca_algo_config_params_min_vals";
	case CA_ALGO_CONFIG_PARAMS_MAX_VALS:
		return "ca_algo_config_params_max_vals";
	default:
		break;
	}

	return "unknown";
}

static const char *cc_algo_get_param_name_by_id(algo_conf_kind_t algo_conf_kind, int param_id)
{
	switch (algo_conf_kind) {
	case CONF_CA_ALGO:
		return get_ca_algo_param_name_by_id(param_id);
	default:
		break;
	}

	return "unknown";
}

static int are_all_params_defined(algo_conf_kind_t algo_conf_kind, boolean_t *params_def, int array_size)
{
	int param_id = 0;

	for (; param_id < array_size; param_id++) {
		if (!params_def[param_id]) {
			yyerror("Param %s is missing",
				cc_algo_get_param_name_by_id(algo_conf_kind, param_id));
			return 0;
		}
	}

	return 1;
}

static void clear_params_defined(boolean_t *params_def, int array_size)
{
	int param_id = 0;

	for (; param_id < array_size; param_id++) {
		params_def[param_id] = FALSE;
	}
}

static void mark_tag_parsed(boolean_t *params_id_array, int array_size,
			    int param_id, const char *param_name)
{
	/* Ignore wrong param ids */
	if (param_id >= array_size)
		return;

	params_id_array[param_id] = TRUE;
}

static inline void mark_ca_algo_tag_parsed(int param_id) {
	mark_tag_parsed(ca_algo_conf_tags_parsed, CA_ALGO_CONFIG_LAST, param_id,
		        get_ca_algo_param_name_by_id(param_id));
}

static int is_tag_parsed(boolean_t *params_id_array, int array_size,
			 int param_id, const char *param_name)
{
	if (param_id >= array_size) {
		yyerror("The param %s has a bigger id %d than the array size",
			param_name, param_id);
		return 1;
	}

	if (params_id_array[param_id]) {
		yyerror("The param %s is already defined. Ignore current parsing "
			"of the param", param_name);
		return 1;
	}

	return 0;

}

static inline int is_ca_algo_tag_parsed(int param_id)
{
	return is_tag_parsed(ca_algo_conf_tags_parsed, CA_ALGO_CONFIG_LAST,
			     param_id, get_ca_algo_param_name_by_id(param_id));
}

static int validate_boolean(uint64_t param, char *name)
{
	if (param != FALSE && param != TRUE) {
		yyerror("%s has invalid value: %llu for a boolean", name, param);
		return 1;
	}

	return 0;
}


static int validate_param_size_by_bits_num(uint64_t param_value,
					   uint8_t bits_num, char *name)
{
	uint64_t max_size = ((uint64_t)1 << bits_num) - 1;

	if (bits_num == 1)
		return validate_boolean(param_value, name);

	if (param_value > max_size) {
		yyerror("The value: %llu of %s is too big, max size is: %llu",
			param_value, name, max_size);
		return 1;
	}

	return 0;
}

static osm_cc_ca_algo_conf_t *create_ca_algo_conf()
{
	osm_cc_ca_algo_conf_t *p_ca_algo_conf = NULL;
	cl_fmap_item_t *p_item;

	p_ca_algo_conf = calloc(1, sizeof(osm_cc_ca_algo_conf_t));
	if (!p_ca_algo_conf) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9E03: "
			"Failed allocating memory for new CA algo conf\n");
		goto Exit;
	}

	memcpy(p_ca_algo_conf->file_name, ppcc_algo_tmp_parser_struct.file_name,
	       strlen(ppcc_algo_tmp_parser_struct.file_name) + 1);

	p_item = cl_fmap_insert(&p_osm->sm.cc_mgr.ca_algo_conf,
				&p_ca_algo_conf->file_name,
				&p_ca_algo_conf->fmap_item);
	if (p_item != &p_ca_algo_conf->fmap_item) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9E04: "
			"Fail to add CA Algo conf, file name: %s of CA Algo "
			"already exist\n", p_ca_algo_conf->file_name);
		free(p_ca_algo_conf);
		p_ca_algo_conf = NULL;
		goto Exit;
	}

	cl_fmap_init(&p_ca_algo_conf->param_name_to_index, ppcc_algo_compare_param_names);
Exit:
	return p_ca_algo_conf;
}

static int ppcc_insert_to_index_map(char *name, int index)
{
	osm_cc_algo_param_index_t *p_algo_param_index = NULL;
	osm_cc_algo_param_index_t *p_map_algo_param_index = NULL;
	cl_fmap_item_t *p_item;
	int ret = 0;

	p_algo_param_index = calloc(1, sizeof(osm_cc_algo_param_index_t));
	if (!p_algo_param_index) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 9E05: Failed to allocate p_algo_param_index\n");
		return 1;
	}

	p_algo_param_index->param_index = index;
	memcpy(p_algo_param_index->param_name, name, sizeof(p_algo_param_index->param_name));

	p_item = cl_fmap_insert(&p_current_ca_algo_conf->param_name_to_index,
				&p_algo_param_index->param_name,
				&p_algo_param_index->fmap_item);
	if (p_item != &p_algo_param_index->fmap_item) {
		p_map_algo_param_index = (osm_cc_algo_param_index_t *) p_item;
		if (p_algo_param_index->param_index != p_map_algo_param_index->param_index) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 9E06: Already in map: our index: %d, index in map: %d, %p\n",
				p_algo_param_index->param_index, p_map_algo_param_index->param_index, p_algo_param_index); 
			ret = 1;
		}

		free(p_algo_param_index);
	}

	return ret;
}

static int ppcc_parse_param_list(int param_id)
{
	osm_cc_algo_param_t *p_algo_param;
	int ret = 0;
	int i;

	if (!p_current_ca_algo_conf) {
		ret = 1;
		goto Exit;
	}

	if (is_ca_algo_tag_parsed(param_id))
		goto Exit;

	mark_ca_algo_tag_parsed(param_id);

	if (!p_current_ca_algo_conf->param_num) {
		p_current_ca_algo_conf->param_num = ppcc_algo_tmp_parser_struct.current_index_in_dictionary;
	} else if (p_current_ca_algo_conf->param_num != ppcc_algo_tmp_parser_struct.current_index_in_dictionary) {
		ret = 1;
		goto Exit;;
	}

	for (i = 0; i < ppcc_algo_tmp_parser_struct.current_index_in_dictionary; i++) {
		p_algo_param = &p_current_ca_algo_conf->algo_param_list[i];
		if (ppcc_insert_to_index_map(ppcc_algo_tmp_parser_struct.dictionary_array[i].str, i)) {
			ret = 1;
			goto Exit;
		}

		switch (param_id) {
		case CA_ALGO_CONFIG_PARAMS_NAME_LIST:
			memcpy(p_algo_param->param_name, ppcc_algo_tmp_parser_struct.dictionary_array[i].str,
			       sizeof(ppcc_algo_tmp_parser_struct.dictionary_array[i].str));
			       p_algo_param->value = ppcc_algo_tmp_parser_struct.dictionary_array[i].value;
			break;
		case CA_ALGO_CONFIG_PARAMS_MIN_VALS:
			p_algo_param->min_value = ppcc_algo_tmp_parser_struct.dictionary_array[i].value;
			break;
		case CA_ALGO_CONFIG_PARAMS_MAX_VALS:
			p_algo_param->max_value = ppcc_algo_tmp_parser_struct.dictionary_array[i].value;
			break;
		default:
			ret = 1;
			goto Exit;
		}
	}

Exit:
	ppcc_algo_tmp_parser_struct.current_index_in_dictionary = 0;

	return ret;
}

int __cc_algo_parser_parse();
int __cc_algo_parser_lex();
void __cc_algo_parser_lex_destroy();

int parse_ppcc_algo_file(struct osm_opensm *osm_ptr, char *file_path, char *file_name)
{
	int res = 0;
	p_osm = osm_ptr;
	__cc_algo_parser_in = NULL;
	parser_tmp_struct_init(file_name);
	cc_algo_line_num = 0;
	cc_algo_column_num = 0;

	if (!file_path) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 9E00: No PPCC algo file provided - Quit PPCC algo parser\n");
		res = 1;
		goto Exit;
	}

	__cc_algo_parser_in = fopen(file_path, "r");
	if (!__cc_algo_parser_in) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 9E01: PPCC algo file specified but not found: %s\n", file_path);
		res = 1;
		goto Exit;
	}

        res =  __cc_algo_parser_parse();
	if (res) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 9E02: PPCC algo file '%s' parsing failed\n", file_path);
		osm_cc_mgr_destroy(&p_osm->sm.cc_mgr);
	} else {
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Congestion control Algo "
			"file parsed successfully\n");
	}

Exit:
	if (__cc_algo_parser_in) {
		__cc_algo_parser_restart(__cc_algo_parser_in);
		fclose(__cc_algo_parser_in);
		__cc_algo_parser_in = NULL;
	}
	clear_val(&__cc_algo_parser_yylval);
	parser_tmp_struct_destroy();
	__cc_algo_parser_lex_destroy();
	return res;
}

%}

%token COMMENT
%token COMMA
%token DOT
%token LEFT_BRACKET
%token RIGHT_BRACKET
%token LEFT_PARENTHESIS
%token RIGHT_PARENTHESIS
%token WORD
%token GROUP
%token EXNUMBER
%token NUMBER




%token CA_ALGO_START
%token CA_ALGO_END
%token ALGO_NAME
%token ALGO_VERSION
%token ALGO_RELEASE_DATE
%token ALGO_DESCRIPTION
%token ALGO_AUTHOR
%token ALGO_SUPPORTED_DEVICES
%token ALGO_VERSION_HASH
%token PPCC_PARAMETER_NAME_LIST
%token PPCC_PARAMETER_MIN_VALS
%token PPCC_PARAMETER_MAX_VALS

%%

cc_policy: configurations;

configurations: /* empty */
		| configurations configuration
		;

comment: COMMENT {};

configuration:    ca_algo
		| word
		| comment
		;


ca_algo: ca_algo_start ca_algo_entries ca_algo_end
	 ;

ca_algo_entries:   /* empty */
		 | ca_algo_entries ca_algo_entry
		 ;

ca_algo_entry:    ca_algo_name
		| ca_algo_version
		| ca_algo_ppcc_parameter_name_list
		| ca_algo_ppcc_parameter_min_vals_list
		| ca_algo_ppcc_parameter_max_vals_list
		| comment
		;

ca_algo_start: CA_ALGO_START {
	p_current_ca_algo_conf = create_ca_algo_conf();
	if (!p_current_ca_algo_conf)
		return 1;

	clear_params_defined(ca_algo_conf_tags_parsed, CA_ALGO_CONFIG_LAST);
};

ca_algo_end: CA_ALGO_END {
	osm_cc_algo_param_t *p_algo_param;
	int i;

	if (!are_all_params_defined(CONF_CA_ALGO, ca_algo_conf_tags_parsed, CA_ALGO_CONFIG_LAST))
		return 1;

	for (i = 0; i < p_current_ca_algo_conf->param_num; i++){
		p_algo_param = &p_current_ca_algo_conf->algo_param_list[i];

		if (p_algo_param->max_value < p_algo_param->min_value) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 9E10: The default min_value of algo param '%s': "
				"%u is bigger then the max value in the allowed range %u\n",
				p_algo_param->param_name, p_algo_param->min_value,
				p_algo_param->max_value);
			return 1;
		}

		if (p_algo_param->value < p_algo_param->min_value ||
		    p_algo_param->max_value < p_algo_param->value) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 9E11: The default value of algo param '%s': "
				"%u is not in the allowed range %u-%u\n",
				p_algo_param->param_name, p_algo_param->value,
				p_algo_param->min_value, p_algo_param->max_value);
			return 1;
		}
	}
};

ca_algo_name: ALGO_NAME WORD {
	if (!p_current_ca_algo_conf)
		return 1;

	if (!is_ca_algo_tag_parsed(CA_ALGO_CONFIG_NAME)) {
		snprintf(p_current_ca_algo_conf->name, sizeof(p_current_ca_algo_conf->name),
			 "%s", __cc_algo_parser_yylval);
		mark_ca_algo_tag_parsed(CA_ALGO_CONFIG_NAME);
	}
};
ca_algo_version: ALGO_VERSION algo_id DOT algo_major DOT algo_minor {
	if (!p_current_ca_algo_conf)
		return 1;

	if (!is_ca_algo_tag_parsed(CA_ALGO_CONFIG_VERSION)) {
		memcpy(&p_current_ca_algo_conf->algo_version, &ppcc_algo_tmp_parser_struct.version,
		       sizeof(ppcc_algo_tmp_parser_struct.version));
		mark_ca_algo_tag_parsed(CA_ALGO_CONFIG_VERSION);
	}
};

algo_id: NUMBER {
	uint64_t algo_id = strtoull(__cc_algo_parser_yylval, NULL, 0);

	if (validate_param_size_by_bits_num(algo_id, PPCC_ALGO_ID_BITS_NUM, "algo_id"))
		return 1;

	ppcc_algo_tmp_parser_struct.version.algo_id = cl_hton16(algo_id);
}

algo_major: NUMBER {
	uint64_t algo_major = strtoull(__cc_algo_parser_yylval, NULL, 0);

	if (validate_param_size_by_bits_num(algo_major, PPCC_ALGO_MAJOR_VERSION_BITS_NUM,
					    "algo_major"))
		return 1;

	ppcc_algo_tmp_parser_struct.version.algo_major_version = algo_major;
}

algo_minor: NUMBER {
	uint64_t algo_minor = strtoull(__cc_algo_parser_yylval, NULL, 0);

	if (validate_param_size_by_bits_num(algo_minor, PPCC_ALGO_MINOR_VERSION_BITS_NUM,
					    "algo_minor"))
		return 1;

	ppcc_algo_tmp_parser_struct.version.algo_minor_version = algo_minor;
}

ca_algo_ppcc_parameter_name_list:  PPCC_PARAMETER_NAME_LIST dictionary {
	if (ppcc_parse_param_list(CA_ALGO_CONFIG_PARAMS_NAME_LIST))
		return 1;
};

ca_algo_ppcc_parameter_min_vals_list:  PPCC_PARAMETER_MIN_VALS dictionary {
	if (ppcc_parse_param_list(CA_ALGO_CONFIG_PARAMS_MIN_VALS))
		return 1;
};

ca_algo_ppcc_parameter_max_vals_list:  PPCC_PARAMETER_MAX_VALS dictionary {
	if (ppcc_parse_param_list(CA_ALGO_CONFIG_PARAMS_MAX_VALS))
		return 1;
};

dictionary: LEFT_BRACKET entries RIGHT_BRACKET
	   ;

entries:  entry
	| entries COMMA entry
	;

entry: LEFT_PARENTHESIS entry_str COMMA entry_value RIGHT_PARENTHESIS {
	osm_ppcc_dictionary_entry_t *dictionary_entry = NULL;

	if (ppcc_algo_tmp_parser_struct.current_index_in_dictionary >= PPCC_ALGO_MAX_PARAMS) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 9E07: Couldn't parse dictionary of PPCC algo parameters, "
			"max number of parameters is %d\n", PPCC_ALGO_MAX_PARAMS);
		return 1;
	}

	dictionary_entry =
	    &ppcc_algo_tmp_parser_struct.dictionary_array[ppcc_algo_tmp_parser_struct.current_index_in_dictionary];

	memcpy(dictionary_entry->str, ppcc_algo_tmp_parser_struct.entry.str, sizeof(ppcc_algo_tmp_parser_struct.entry.str));
	dictionary_entry->value = ppcc_algo_tmp_parser_struct.entry.value;
	ppcc_algo_tmp_parser_struct.current_index_in_dictionary++;
};

entry_str: WORD {
	memcpy(ppcc_algo_tmp_parser_struct.entry.str, __cc_algo_parser_yylval,
	       strlen(__cc_algo_parser_yylval) + 1);
};

entry_value: NUMBER {
	uint64_t tmp = strtoull(__cc_algo_parser_yylval, NULL, 0);

	if (tmp > UINT32_MAX)
		yyerror("The value %llu is to big. The max value is %u\n", UINT32_MAX);

	ppcc_algo_tmp_parser_struct.entry.value = tmp;
};

word: WORD {
	OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9E08: Unexpected "
		"input |%.200s|, Aborting parsing!\n", __cc_algo_parser_yylval);
	return 1;
};
