/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2021-2022 Nvidia, Inc. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of Ford forlkenson max flow algorithm.
 * This file implements the algorithm. Currently only for trees with
 * the same speeds on all of the links.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_MAX_FLOW_ALGORITHM_C
#include <opensm/osm_max_flow_algorithm.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>

static inline boolean_t flow_is_residual(flow_edge_t *p_edge)
{
	return (p_edge->capacity == 0);
}

static inline int flow_get_remaining_capacity(flow_edge_t *p_edge)
{
	return (p_edge->capacity - p_edge->flow);
}

static inline void flow_augment(flow_edge_t *p_edge, int bottle_neck, flow_edge_t *edges)
{
	flow_edge_t *p_residual_edge = &edges[p_edge->residual_edge_index];

	p_edge->flow += bottle_neck;
	p_residual_edge->flow -= bottle_neck;
}

static flow_edge_t *flow_add_new_edge(flow_vertex_t *vertex, flow_edge_t *edges)
{
	vertex->end_edges_index++;
	return &edges[vertex->end_edges_index - 1];
}

static void flow_init_edge(flow_edge_t *p_edge, int index, int port_num,
			   int adj_index, int adj_port_num, int capacity)
{
	p_edge->index = index;
	p_edge->port_num = port_num;
	p_edge->adj_index = adj_index;
	p_edge->adj_port_num = adj_port_num;
	p_edge->capacity = capacity;
	p_edge->flow = 0;
}

static void flow_add_edge(flow_vertex_t *vertices, flow_edge_t *edges, int index, int port_num,
			  int adj_index, int adj_port_num, int capacity)
{
	flow_edge_t *p_edge = flow_add_new_edge(&vertices[index], edges);
	flow_edge_t *p_residual_edge = flow_add_new_edge(&vertices[adj_index], edges);

	flow_init_edge(p_edge, index, port_num,
		       adj_index, adj_port_num, capacity);
	flow_init_edge(p_residual_edge, adj_index, adj_port_num,
		       index, port_num, capacity);

	p_edge->residual_edge_index = vertices[adj_index].end_edges_index - 1;
	p_residual_edge->residual_edge_index = vertices[index].end_edges_index - 1;
}

static void destroy_count_sort_array(flow_graph_t *p_graph)
{
	int i;
	int array_size = p_graph->count_sort_array_size;
	cl_list_t *count_sort = p_graph->count_sort;

	if (!count_sort)
		return;

	for (i = 0; i < array_size; i++) {
		cl_list_t *p_list = &count_sort[i];

		cl_list_remove_all(p_list);
		cl_list_destroy(p_list);
	}

	free(p_graph->count_sort);
	p_graph->count_sort = NULL;
}

static void flow_destroy_port_groups_array(flow_graph_t *p_graph)
{
	int i, j;
	flow_port_groups_t *src_port_groups = p_graph->src_port_groups;
	flow_port_group_t *port_groups;
	flow_port_group_t *port_group;

	if (!src_port_groups)
		return;

	for (i = 0; i < p_graph->vertices_num; i++) {
		port_groups = src_port_groups[i].port_groups;
		if (!port_groups)
			continue;

		for (j = 0; j < p_graph->vertices_num; j++) {
			port_group = &port_groups[j];

			if (port_group->ports) {
				free(port_group->ports);
				port_group->num_ports = 0;
			}
		}
		free(port_groups);
	}

	free(src_port_groups);
	p_graph->src_port_groups = NULL;
}

void osm_flow_destroy_graph(flow_graph_t **pp_graph)
{
	flow_graph_t *p_graph = *pp_graph;
	flow_vertex_t *vertices;
	void **prev_privs;
	osm_switch_t *p_sw;
	int i;

	if (!p_graph)
		return;

	if (p_graph->prev_privs) {
		vertices = p_graph->vertices;
		prev_privs = p_graph->prev_privs;
		for (i = 0; i < p_graph->vertices_num; i++) {
			p_sw = vertices[i].p_sw;
			p_sw->priv = prev_privs[i];
		}

		free(p_graph->prev_privs);
	}

	if (p_graph->src_port_groups)
		flow_destroy_port_groups_array(p_graph);

	free(p_graph->visited);
	free(p_graph->dirty);
	free(p_graph->vertices);
	free(p_graph->edges);
	destroy_count_sort_array(p_graph);
	free(p_graph);
	*pp_graph = NULL;
}

static flow_graph_t *flow_allocate_graph(osm_subn_t *p_subn,
					 boolean_t save_prev_privs)
{
	flow_graph_t *p_graph = NULL;
	flow_vertex_t *vertices = NULL;
	int *visited = NULL;
	boolean_t *dirty = NULL;
	void **prev_privs = NULL;
	unsigned int num_vertices = 0;
	int i;

	num_vertices = cl_qmap_count(&p_subn->sw_guid_tbl);
	if (num_vertices == 0) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9735: Cannot route topology without switches\n");
		goto Exit;
	}

	p_graph = (flow_graph_t *)calloc(1, sizeof(flow_graph_t));
	if (!p_graph) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9736: Cannot allocate memory for graph\n");
		goto Exit;
	}

	/* Create vertices array. */
	vertices = (flow_vertex_t *)malloc(num_vertices * sizeof(flow_vertex_t));
	if (!vertices) {
		free(p_graph);
		p_graph = NULL;
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9737: Cannot allocate memory for vertices\n");
		goto Exit;
	}

	for (i = 0; i < num_vertices; i++) {
		flow_vertex_t *p_vertex = &vertices[i];

		p_vertex->rank = 0;
		p_vertex->start_edges_index = 0;
		p_vertex->end_edges_index = 0;
		p_vertex->calculate_flow = FALSE;
	}

	visited = (int *)calloc(num_vertices, sizeof(int));
	if (!visited) {
		free(vertices);
		free(p_graph);
		p_graph = NULL;
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9738: Cannot allocate memory for visited array\n");
		goto Exit;
	}

	dirty = (int *)calloc(num_vertices, sizeof(int));
	if (!dirty) {
		free(visited);
		free(vertices);
		free(p_graph);
		p_graph = NULL;
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9739: Cannot allocate memory for dirty array\n");
		goto Exit;
	}

	if (save_prev_privs) {
		prev_privs = (void **)calloc(num_vertices, sizeof(void *));
		if (!prev_privs) {
			free(dirty);
			free(visited);
			free(vertices);
			free(p_graph);
			p_graph = NULL;
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 9740: Cannot allocate memory for prev_privs array\n");
			goto Exit;
		}
		p_graph->prev_privs = prev_privs;
	}

	p_graph->vertices = vertices;
	p_graph->visited = visited;
	p_graph->dirty = dirty;
	p_graph->vertices_num = num_vertices;
	p_graph->p_subn = p_subn;
	p_graph->count_sort = NULL;

Exit:
	return p_graph;
}

static flow_edge_t *flow_init_edges_array(flow_graph_t *p_graph, unsigned int num_edges)
{
	flow_edge_t *edges;

	edges = (flow_edge_t *)malloc(num_edges * sizeof(flow_edge_t));
	if (!edges) {
		OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9743: Cannot allocate memory for edges\n");
		goto Exit;
	}

	p_graph->num_edges = num_edges;
	p_graph->edges = edges;
Exit:
	return p_graph->edges;
}

flow_graph_t *osm_flow_build_graph(osm_subn_t *p_subn)
{
	unsigned int *sw_lid_to_vertex_mapping;
	cl_map_item_t *item;
	osm_switch_t *p_sw, *p_rem_sw;
	osm_node_t *p_node, *p_rem_node;
	flow_vertex_t *vertices = NULL;
	flow_vertex_t *p_vertex, *p_adj_vertex;
	unsigned int num_vertices;
	unsigned int num_edges = 0;
	unsigned int index;
	uint8_t port_num, rem_port_num;
	flow_graph_t *p_graph = NULL;
	int *visited;
	osm_physp_t *p_physp;
	void **prev_privs;
	int flow_calculations_num = 0;
	uint16_t sw_lid;

	p_graph = flow_allocate_graph(p_subn, TRUE);
	if (!p_graph)
		goto Exit;

	num_vertices = p_graph->vertices_num;
	visited = p_graph->visited;
	vertices = p_graph->vertices;
	sw_lid_to_vertex_mapping = p_graph->sw_lid_to_vertex_mapping;
	prev_privs = p_graph->prev_privs;

	index = 0;
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		p_vertex = &vertices[index];
		p_vertex->p_sw = p_sw;
		prev_privs[index] = p_sw->priv;
		p_sw->priv = p_vertex;
		p_vertex->index = index;
		index++;
	}

	for (index = 0; index < num_vertices; index++) {
		p_vertex = &vertices[index];
		p_vertex->start_edges_index = num_edges;
		p_vertex->end_edges_index = num_edges;
		p_sw = p_vertex->p_sw;
		p_node = p_sw->p_node;
		sw_lid = osm_node_get_base_lid_ho(p_node, 0);
		sw_lid_to_vertex_mapping[sw_lid] = index;
		for (port_num = 1; port_num < p_sw->num_ports; port_num++) {
			p_rem_node = osm_node_get_remote_node(p_node, port_num,
							      &rem_port_num);
			if (!p_rem_node || p_rem_node == p_node || !p_rem_node->sw)
				continue;

			num_edges++;
		}
	}

	flow_init_edges_array(p_graph, num_edges);

	for (index = 0; index < num_vertices; index++) {
		boolean_t hca_exists = FALSE;
		boolean_t is_leaf = TRUE;
		int num_edges;

		p_vertex = &vertices[index];
		visited[index] = 1;
		p_sw = p_vertex->p_sw;
		p_node = p_sw->p_node;
		for (port_num = 1; port_num < p_sw->num_ports; port_num++) {
			int adj_index;

			p_rem_node = osm_node_get_remote_node(p_node, port_num,
							      &rem_port_num);
			if (!p_rem_node || p_rem_node == p_node)
				continue;

			if (osm_node_get_type(p_rem_node)  == IB_NODE_TYPE_ROUTER)
				continue;

			if (!p_rem_node->sw) {
				p_physp = osm_node_get_physp_ptr(p_node, port_num);
				if (!osm_physp_get_special_port_type(p_physp->p_remote_physp))
					hca_exists = TRUE;

				continue;
			}

			p_rem_sw = p_rem_node->sw;
			if (p_sw->rank < p_rem_sw->rank)
				is_leaf = FALSE;

			p_adj_vertex = (flow_vertex_t *)p_rem_sw->priv;
			adj_index = p_adj_vertex->index;
			if (!visited[adj_index])
				flow_add_edge(vertices, p_graph->edges, index, port_num,
					      adj_index, rem_port_num, 1);
		}

		num_edges = p_vertex->end_edges_index - p_vertex->start_edges_index;
		if (is_leaf && hca_exists && num_edges) {
			flow_calculations_num++;
			p_vertex->calculate_flow = TRUE;
		}
	}

	/*
	 * To know what is the maximum number of times a port can be part of an
	 * augmented path in one max flow calculation need to calculate:
	 * max_augment_port_picks = max(c1,c2,...) / gcd(c1,c2,...)
	 * c1,c2... stands for the capacity/speed of the links.
	 * Need to multiply the flow_calculations_num by max_augment_port_picks
	 * to know the maximum number a port can be picked in all
	 * the calculations from a specific source,
	 * Currently, the algorithm supports only topologies with the same
	 * speed on all the links, so the value of max_augment_port_picks is 1.
	 */
	p_graph->count_sort_array_size = flow_calculations_num;

Exit:
	return p_graph;
}

static cl_status_t flow_init_count_sort_array(flow_graph_t *p_graph)
{
	int i;
	cl_status_t status = CL_SUCCESS;

	p_graph->count_sort = (cl_list_t *)calloc(p_graph->count_sort_array_size,
						  sizeof(cl_list_t));
	if (!p_graph->count_sort) {
		OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 3339: Cannot allocate memory for count_sort array\n");
		status = CL_INSUFFICIENT_RESOURCES;
		goto Exit;
	}

	for (i = 0; i < p_graph->count_sort_array_size; i++)
		cl_list_init(&p_graph->count_sort[i], 10);

Exit:
	return status;
}

flow_graph_t *osm_flow_copy_graph(flow_graph_t *p_graph)
{
	flow_graph_t *p_new_graph = flow_allocate_graph(p_graph->p_subn, FALSE);
	int num_vertices, index;
	flow_vertex_t *new_vertices, *vertices;
	flow_vertex_t *p_new_vertex;
	unsigned int *new_sw_lid_to_vertex_mapping;
	flow_edge_t *edges = p_graph->edges;
	flow_edge_t *new_edges = NULL;
	int *new_visited;
	u_int16_t sw_lid;

	if (!p_new_graph)
		goto Exit;

	p_new_graph->count_sort_array_size = p_graph->count_sort_array_size;
	vertices = p_graph->vertices;
	num_vertices = p_new_graph->vertices_num;
	new_vertices = p_new_graph->vertices;
	new_visited = p_new_graph->visited;
	new_sw_lid_to_vertex_mapping = p_new_graph->sw_lid_to_vertex_mapping;

	flow_init_edges_array(p_new_graph, p_graph->num_edges);
	new_edges = p_new_graph->edges;
	memcpy(new_edges, edges, sizeof(flow_edge_t) * p_graph->num_edges);
	memcpy(new_vertices, vertices, sizeof(flow_vertex_t) * num_vertices);
	for (index = 0; index < num_vertices; index++) {
		p_new_vertex = &new_vertices[index];
		new_visited[index] = 1;

		sw_lid = osm_node_get_base_lid_ho(p_new_vertex->p_sw->p_node, 0);
		new_sw_lid_to_vertex_mapping[sw_lid] = p_new_vertex->index;
	}

	if (flow_init_count_sort_array(p_new_graph) != CL_SUCCESS)
		osm_flow_destroy_graph(&p_new_graph);

Exit:
	return p_new_graph;
}

static int _flow_dfs_solve(flow_graph_t *p_graph, flow_vertex_t *p_vertex,
			   int flow, boolean_t is_src_vertex, int *start_idx)
{
	flow_vertex_t *vertices = p_graph->vertices;
	flow_edge_t *edges = p_graph->edges;
	uint16_t dst_lid_ho = p_graph->dst_lid_ho;
	int *visited = p_graph->visited;
	boolean_t *dirty = p_graph->dirty;
	int min_flow, bottle_neck, adj_index;
	int remaining_capacity;
	unsigned int i;
	flow_edge_t *p_edge = NULL;
	flow_vertex_t *p_adj_vertex;
	uint16_t min_hops_vertex;
	uint8_t **hops;
	cl_list_iterator_t next_list_iter = NULL;
	cl_list_iterator_t cur_list_iter = NULL;
	cl_list_t *p_list = NULL;
	cl_list_t *p_tmp_list = NULL;
	int j = *start_idx;
	int count_sort_array_size = p_graph->count_sort_array_size;

	if (p_vertex == p_graph->p_dst_vertex)
		return flow;

	visited[p_vertex->index] = p_graph->visited_token;
	min_hops_vertex = osm_switch_get_least_hops(p_vertex->p_sw, dst_lid_ho);
	hops = p_vertex->p_sw->hops;

	if (is_src_vertex) {
		p_list = &p_graph->count_sort[j];
		next_list_iter = cl_list_head(p_list);
		cur_list_iter = next_list_iter;
	}

	for (i = p_vertex->start_edges_index; i < p_vertex->end_edges_index; i++) {
		p_edge = NULL;
		if (is_src_vertex) {
			while (j < count_sort_array_size) {
				if (next_list_iter != cl_list_end(p_list)) {
					p_edge = (flow_edge_t *) cl_list_obj(next_list_iter);
					cur_list_iter = next_list_iter;
					next_list_iter = cl_list_next(cur_list_iter);
					break;
				} else {
					if ((j == *start_idx) && !cl_list_count(p_list))
						*start_idx = j + 1;

					j++;
					if (j < count_sort_array_size) {
						p_list = &p_graph->count_sort[j];
						next_list_iter = cl_list_head(p_list);
					}
				}
			}

			if (!p_edge) {
				OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 3340: edge missing in count sort array\n");
				return 0;
			}
		} else {
			p_edge = &edges[i];
		}

		p_adj_vertex = &vertices[p_edge->adj_index];
		adj_index = p_adj_vertex->index;

		remaining_capacity = flow_get_remaining_capacity(p_edge);
		if (remaining_capacity && (visited[adj_index] != p_graph->visited_token) &&
		    (flow_is_residual(p_edge) || (min_hops_vertex == hops[dst_lid_ho][p_edge->port_num]))) {
			min_flow = remaining_capacity > flow ? flow : remaining_capacity;
			bottle_neck = _flow_dfs_solve(p_graph, p_adj_vertex, min_flow, FALSE, start_idx);

			if (bottle_neck > 0) {
				flow_augment(p_edge, bottle_neck, edges);
				dirty[adj_index] = TRUE;
				if (is_src_vertex) {
					cl_list_remove_item(p_list, cur_list_iter);
					p_tmp_list = &p_graph->count_sort[j + bottle_neck];
					cl_list_insert_tail(p_tmp_list, p_edge);
				}

				return bottle_neck;
			}
		}
	}

	return 0;
}


static int flow_dfs_solve(flow_graph_t *p_graph, flow_vertex_t *p_src_vertex, int *start_idx)
{
	if (p_src_vertex == p_graph->p_dst_vertex)
		return 0;

	return _flow_dfs_solve(p_graph, p_src_vertex, INT32_MAX, TRUE, start_idx);
}

void flow_edge_print(flow_graph_t *p_graph, flow_edge_t *p_edge)
{
	flow_vertex_t *vertices = p_graph->vertices;
	flow_vertex_t *p_vertex = &vertices[p_edge->index];
	flow_vertex_t *p_adj_vertex = &vertices[p_edge->adj_index];

	OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
		"Edge 0x%016" PRIx64 "[%d] --> 0x%016" PRIx64
		"[%d], flow = %d, capacity = %d, is_residual = %d\n",
		osm_node_get_node_guid_ho(p_vertex->p_sw->p_node), p_edge->port_num,
		osm_node_get_node_guid_ho(p_adj_vertex->p_sw->p_node),
		p_edge->adj_port_num, p_edge->flow, p_edge->capacity, flow_is_residual(p_edge));
}

void flow_print_edges(flow_graph_t *p_graph)
{
	flow_vertex_t *vertices = p_graph->vertices;
	flow_edge_t *edges = p_graph->edges;
	flow_vertex_t *p_vertex;
	int index, e;

	for (index = 0; index < p_graph->vertices_num; index++) {
		p_vertex = &vertices[index];

		for (e = p_vertex->start_edges_index; e < p_vertex->end_edges_index; e++)
			flow_edge_print(p_graph, &edges[e]);
	}
}

int osm_flow_max_flow_solve(flow_graph_t *p_graph, flow_vertex_t *p_src_vertex,
			    flow_vertex_t *p_dst_vertex, int *start_idx)
{
	int f;
	int max_flow = 0;
	uint16_t dst_lid_ho;
	uint16_t hops_vertex;

	p_graph->visited_token = 2;
	p_graph->p_dst_vertex = p_dst_vertex;

	dst_lid_ho = osm_node_get_base_lid_ho(p_graph->p_dst_vertex->p_sw->p_node, 0);
	hops_vertex = osm_switch_get_least_hops(p_src_vertex->p_sw, dst_lid_ho);
	if (hops_vertex == OSM_NO_PATH) {
		OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_VERBOSE,
			"No path between representative src SW: 0x%016" PRIx64
			", dst SW: 0x%016" PRIx64 "\n",
			osm_node_get_node_guid_ho(p_src_vertex->p_sw->p_node),
			osm_node_get_node_guid_ho(p_dst_vertex->p_sw->p_node));

		return 0;
	}

	p_graph->dst_lid_ho = dst_lid_ho;

	for (f = flow_dfs_solve(p_graph, p_src_vertex, start_idx);
	     f != 0; f = flow_dfs_solve(p_graph, p_src_vertex, start_idx)) {
		p_graph->visited_token++;
		max_flow += f;
	}

	OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_DEBUG, "Src SW: "
		"0x%016" PRIx64 ", Dst SW: 0x%016" PRIx64
		", visited_token: %d, max_flow: %d\n",
		osm_node_get_node_guid_ho(p_src_vertex->p_sw->p_node),
		osm_node_get_node_guid_ho(p_dst_vertex->p_sw->p_node),
		p_graph->visited_token, max_flow);

	return max_flow;
}



int osm_flow_update_edge_directions(flow_graph_t *p_graph, flow_vertex_t *p_src_vertex)
{
	int result = 0;
	unsigned int queue_in = 0;
	unsigned int queue_out = 0;
	osm_log_t *p_log = &p_graph->p_subn->p_osm->log;
	flow_vertex_t **queue = NULL;
	flow_vertex_t *vertices = p_graph->vertices;
	flow_edge_t *edges = p_graph->edges;
	int *visited = p_graph->visited;
	flow_vertex_t *p_vertex, *p_adj_vertex;
	flow_edge_t *p_edge;
	int visited_token;
	int edge_index, index, adj_index;

	queue = (flow_vertex_t **)calloc(p_graph->vertices_num, sizeof(flow_vertex_t*));
	if (!queue) {
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR 9741: Cannot allocate memory for routing"
			" engine\n");
		result = 1;
		goto Exit;
	}

	p_graph->visited_token = 1;
	visited_token = p_graph->visited_token;
	memset(visited, 0, sizeof(int) * p_graph->vertices_num);
	for (index = 0; index < p_graph->vertices_num; index++)
		vertices[index].rank = 0;

	queue[queue_in++] = p_src_vertex;
	visited[p_src_vertex->index] = p_graph->visited_token;

	while (queue_out < queue_in) {
		p_vertex = queue[queue_out++];

		/* Add adjacent vertices to queue. */
		for (edge_index = p_vertex->start_edges_index; edge_index < p_vertex->end_edges_index; edge_index++) {
			p_edge = &edges[edge_index];
			p_adj_vertex = &vertices[p_edge->adj_index];
			adj_index = p_adj_vertex->index;

			if (visited[adj_index] != visited_token) {
				visited[adj_index] = visited_token;
				p_adj_vertex->rank = p_vertex->rank + 1;
				queue[queue_in++] = p_adj_vertex;
			}

			p_edge->capacity = p_adj_vertex->rank > p_vertex->rank ? 1 : 0;
		}
	}

	free(queue);
Exit:
	return result;
}

void osm_flow_init_flow_and_visited(flow_graph_t *p_graph)
{
	flow_vertex_t *vertices = p_graph->vertices;
	flow_edge_t *edges = p_graph->edges;
	boolean_t *dirty = p_graph->dirty;
	flow_vertex_t *p_vertex;
	flow_edge_t *p_edge, *p_residual_edge;
	int index, e;

	memset(p_graph->visited, 0, sizeof(int) * p_graph->vertices_num);
	for (index = 0; index < p_graph->vertices_num; index++) {
		if (!dirty[index])
			continue;

		p_vertex = &vertices[index];
		dirty[index] = FALSE;
		for (e = p_vertex->start_edges_index; e < p_vertex->end_edges_index; e++) {
			p_edge = &edges[e];
			p_residual_edge = &edges[p_edge->residual_edge_index];

			p_edge->flow = 0;
			p_residual_edge->flow = 0;
		}
	}
}

typedef struct vertex_index_to_port_ {
	int adj_index;
	int port_num;
} vertex_index_to_port_t;

int cmpfunc (const void * lhs, const void * rhs) {
	vertex_index_to_port_t *p_left = (vertex_index_to_port_t *) lhs;
	vertex_index_to_port_t *p_right = (vertex_index_to_port_t *) rhs;

	if (p_left->adj_index == p_right->adj_index)
		return (p_left->port_num - p_right->port_num);

	return (p_left->adj_index - p_right->adj_index);
}

typedef struct key_to_vertex_ {
	char key[2048];
	int vertex_index;
	uint8_t ports[FLOW_MAX_SUPPORTED_PORTS];
} key_to_vertex_t;

static int key_to_vertex_cmp_func(const void *lhs, const void* rhs)
{
	key_to_vertex_t *p_left = (key_to_vertex_t *)lhs;
	key_to_vertex_t *p_right = (key_to_vertex_t *)rhs;

	return strcmp(p_left->key, p_right->key);
}

static cl_status_t flow_init_port_groups_array(flow_graph_t *p_graph)
{
	int i;
	cl_status_t status = CL_SUCCESS;

	p_graph->src_port_groups = (flow_port_groups_t *)calloc(p_graph->vertices_num, sizeof(flow_port_groups_t));
	if (!p_graph->src_port_groups) {
		OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9746: Cannot allocate memory for src_port_groups array\n");
		status = CL_INSUFFICIENT_RESOURCES;
		goto Exit;
	}

	for (i = 0; i < p_graph->vertices_num; i++) {
		p_graph->src_port_groups[i].port_groups =
		    (flow_port_group_t * ) calloc(p_graph->vertices_num,
						  sizeof(flow_port_group_t));
		if (!p_graph->src_port_groups[i].port_groups) {
			OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 9747: Cannot allocate memory for port_groups array\n");
			status = CL_INSUFFICIENT_RESOURCES;
			goto Exit;
		}
	}

Exit:
	return status;
}

static void init_representative_vertex_index(flow_graph_t *p_graph, key_to_vertex_t *current_key,
					     key_to_vertex_t *current_representative_key,
					     key_to_vertex_t *current_second_representative_key)
{
	flow_port_groups_t *src_port_groups = &p_graph->src_port_groups[current_key->vertex_index];
	uint8_t port_of_first_representative, port_of_second_representative, port_of_current_key;
	int i;

	src_port_groups->vertex_index_of_representative = current_representative_key->vertex_index;
	src_port_groups->vertex_index_of_second_representative = current_second_representative_key->vertex_index;

	for (i = 0; i < FLOW_MAX_SUPPORTED_PORTS; i++) {
		port_of_first_representative = current_representative_key->ports[i];
		port_of_second_representative = current_second_representative_key->ports[i];
		port_of_current_key = current_key->ports[i];
		src_port_groups->representative_port_map[port_of_first_representative] = port_of_current_key;
		src_port_groups->second_representative_port_map[port_of_second_representative] = port_of_current_key;
	}
}

void flow_choose_groups(flow_graph_t *p_graph)
{
	flow_vertex_t *vertices = p_graph->vertices;
	flow_edge_t *edges = p_graph->edges;
	flow_edge_t *p_edge;
	int i, e, n, j, num_edges, num_sources, num_in_current_group;
	int key_index = 0;
	vertex_index_to_port_t *adj_indexes;
	key_to_vertex_t *current_key = NULL;
	key_to_vertex_t *current_second_key = NULL;
	key_to_vertex_t *keys_indexes = (key_to_vertex_t *) calloc(p_graph->count_sort_array_size, sizeof(key_to_vertex_t));
	if (!keys_indexes) {
		OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9745: Cannot allocate memory for array of key indexes\n");
		goto Exit;
	}

	if (flow_init_port_groups_array(p_graph))
		goto Exit;

	for (i = 0; i < p_graph->vertices_num; i++) {
		char *key = keys_indexes[key_index].key;
		flow_vertex_t *p_vertex = &vertices[i];

		if (!p_vertex->calculate_flow)
			continue;

		keys_indexes[key_index].vertex_index = p_vertex->index;
		num_edges = p_vertex->end_edges_index - p_vertex->start_edges_index;
		adj_indexes = (vertex_index_to_port_t *) malloc(sizeof(vertex_index_to_port_t) * num_edges);
		if (!adj_indexes) {
			OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 9748: Cannot allocate memory for adj_indexes key array\n");
			goto Exit;
		}

		j = 0;
		for (e = p_vertex->start_edges_index; e < p_vertex->end_edges_index; e++) {
			p_edge = &edges[e];
			adj_indexes[j].adj_index = p_edge->adj_index;
			adj_indexes[j].port_num = p_edge->port_num;
			j++;
		}
		qsort(adj_indexes, num_edges, sizeof(vertex_index_to_port_t), cmpfunc);

		n = 0;
		for (j = 0; j < num_edges; j++) {
			keys_indexes[key_index].ports[j] = adj_indexes[j].port_num;
			if (j == 0)
				n = sprintf(key, "%d", adj_indexes[j].adj_index);
			else
				n += snprintf(key + n, sizeof(keys_indexes[key_index].key) - n,
					      ",%d", adj_indexes[j].adj_index);
		}
		key_index++;
		free(adj_indexes);
	}

	qsort(keys_indexes, p_graph->count_sort_array_size, sizeof(key_to_vertex_t), key_to_vertex_cmp_func);

	if (p_graph->count_sort_array_size < 2)
		goto Exit;

	current_key = &keys_indexes[0];
	current_second_key = &keys_indexes[0];
	init_representative_vertex_index(p_graph, current_key,
					 current_key, current_second_key);
	OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_DEBUG,
		"%s, rep 1: %d, rep 2: %d\n", current_key->key,
		current_key->vertex_index, current_second_key->vertex_index);
	num_sources = 1;
	num_in_current_group = 1;
	for (i = 1; i < p_graph->count_sort_array_size; i++) {
		if (strcmp(current_key->key, keys_indexes[i].key)) {
			current_key = &keys_indexes[i];
			current_second_key = &keys_indexes[i];
			num_sources++;

			num_in_current_group = 1;
		} else {
			num_in_current_group++;
			if (num_in_current_group == 2)
				current_second_key = &keys_indexes[i];
		}

		init_representative_vertex_index(p_graph, &keys_indexes[i],
						 current_key, current_second_key);
		OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_DEBUG,
			"%s, rep 1: %d, rep 2: %d\n", keys_indexes[i].key,
			current_key->vertex_index, current_second_key->vertex_index);
	}

	OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_DEBUG,
		"flow groups num: %d\n", num_sources);
Exit:
	free(keys_indexes);
	return;
}


static void flow_run_src_to_dst_flow(flow_graph_t *p_graph,
				     flow_vertex_t *p_src_vertex,
				     int dst_index,
				     flow_port_groups_t *src_port_groups,
				     int *p_start_index)
{
	flow_vertex_t *p_dst_vertex = &p_graph->vertices[dst_index];
	flow_edge_t *edges = p_graph->edges;
	flow_edge_t *p_edge;
	uint8_t *ports;
	int num_ports = 0;
	int e;

	if ((p_src_vertex == p_dst_vertex) || !p_dst_vertex->calculate_flow)
		goto Exit;

	osm_flow_init_flow_and_visited(p_graph);
	osm_flow_max_flow_solve(p_graph, p_src_vertex, p_dst_vertex, p_start_index);
	ports = (uint8_t *) malloc(p_src_vertex->end_edges_index - p_src_vertex->start_edges_index);
	if (!ports) {
		OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9749: Cannot allocate memory for ports in the flow port_group array\n");
		goto Exit;
	}
	for (e = p_src_vertex->start_edges_index; e < p_src_vertex->end_edges_index; e++) {
		p_edge = &edges[e];
		if (p_edge->flow == 1) {
			ports[num_ports] = p_edge->port_num;
			num_ports++;
		}
	}
	src_port_groups->port_groups[dst_index].ports = ports;
	src_port_groups->port_groups[dst_index].num_ports = num_ports;
Exit:
	return;
}

/* Note: Currently this function supports only batches of size 1 */
static int flow_run_src_to_all_dst_flow(void **task, unsigned task_len, void *context,
					unsigned thread_id)
{
	flow_graph_t *p_ctx_graph = (flow_graph_t *)context;
	flow_vertex_t *p_ctx_src_vertex = *((flow_vertex_t **)task);
	flow_vertex_t *p_src_vertex;
	flow_graph_t *p_graph = NULL;
	int start_index = 0;
	int dst_index, i;
	flow_port_groups_t *src_port_groups;

	p_graph = osm_flow_copy_graph(p_ctx_graph);
	if (!p_graph)
		return 1;

	p_src_vertex = &p_graph->vertices[p_ctx_src_vertex->index];
	osm_flow_update_edge_directions(p_graph, p_src_vertex);
	for (i = p_src_vertex->start_edges_index; i < p_src_vertex->end_edges_index; i++) {
		cl_list_insert_tail(&p_graph->count_sort[0], &p_graph->edges[i]);
	}

	src_port_groups = &p_ctx_graph->src_port_groups[p_src_vertex->index];
	if (src_port_groups->vertex_index_of_representative == p_src_vertex->index) {
		for (dst_index = 0; dst_index < p_graph->vertices_num; dst_index++) {
			flow_run_src_to_dst_flow(p_graph, p_src_vertex, dst_index,
						 src_port_groups, &start_index);
		}
	} else if (src_port_groups->vertex_index_of_second_representative == p_src_vertex->index) {
		dst_index = src_port_groups->vertex_index_of_representative;
		flow_run_src_to_dst_flow(p_graph, p_src_vertex, dst_index, src_port_groups, &start_index);
	}

	osm_flow_destroy_graph(&p_graph);

	return 0;
}

void osm_flow_calculate_parallel_flow(flow_graph_t *p_graph)
{
	flow_vertex_t *vertices = p_graph->vertices;
	flow_port_groups_t *src_port_groups;
	flow_vertex_t *p_vertex;
	void **adr_vertices;
	int num_sources = 0;
	int leafs_num = 0;
	int index = 0;
	int i;

	flow_choose_groups(p_graph);
	for (i = 0; i < p_graph->vertices_num; i++) {
		p_vertex = &vertices[i];
		src_port_groups = &p_graph->src_port_groups[p_vertex->index];
		if (!p_vertex->calculate_flow)
			continue;

		leafs_num++;
		if ((src_port_groups->vertex_index_of_representative != p_vertex->index) &&
		    (src_port_groups->vertex_index_of_second_representative != p_vertex->index))
			continue;

		num_sources++;
	}

	if (!num_sources) {
		if (leafs_num > 1) {
			OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 9750: No sources found to calculate flow from\n");
		}
		goto Exit;
	}

	adr_vertices = (void**) malloc(sizeof(void *) * num_sources);
	if (!adr_vertices) {
		OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9744: Cannot allocate memory for array of vertices\n");
		goto Exit;
	}

	for (i = 0; i < p_graph->vertices_num && index < num_sources; i++) {
		p_vertex = &vertices[i];
		src_port_groups = &p_graph->src_port_groups[p_vertex->index];
		if (!p_vertex->calculate_flow)
			continue;

		if ((src_port_groups->vertex_index_of_representative != p_vertex->index) &&
		    (src_port_groups->vertex_index_of_second_representative != p_vertex->index))
			continue;

		adr_vertices[index] = p_vertex;
		index++;
	}

	/* Note flow_run_src_to_all_dst_flow supports only batches of size 1 */
	cl_mp_thread_pool_array_apply(&p_graph->p_subn->p_osm->sm.ucast_mgr.thread_pool,
				      adr_vertices, num_sources,
				      CL_MP_THREAD_POOL_BATCH_PER_TASK,
				      flow_run_src_to_all_dst_flow,
				      p_graph);
	free(adr_vertices);
Exit:
	return;
}
