/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Main entry point into routing subsystem.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTING_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_routing.h>
#include <opensm/osm_advanced_routing.h>

/* -------------------------------------------------------------------------------------------------
 *
 * Routing engine private definitions.
 */

#define	LOADABLE_ROUTING_ENGINE_TYPE_BASE	128

static cl_qlist_t       osm_routing_engine_list;

typedef struct {
        cl_list_item_t  		list_item;
        char      			*name;
        void            		*setup;
        osm_routing_engine_type_t	type;
} re_module_t;


#define foreach_routing_engine(P_LIST, P_MODULE)                                        \
	for (P_MODULE  = (re_module_t *)cl_qlist_head(P_LIST);                          \
	     P_MODULE != (re_module_t *)cl_qlist_end(P_LIST);                           \
	     P_MODULE  = (re_module_t *)cl_qlist_next((cl_list_item_t *)P_MODULE))


static re_module_t *routing_engine_find(IN cl_qlist_t *p_list, IN const char *name)
{
	re_module_t     *p_module;

	foreach_routing_engine(p_list, p_module) {
		if (strcasecmp(name, p_module->name) == 0) {
			return p_module;
		}
	}

	return NULL;
}

static f_re_setup_t osm_routing_engine_setup(IN const char *name)
{
	re_module_t     *p_module;

	if ((name == NULL) || (strcmp(name, "none") == 0)) {
		name = "minhop";
	}

	if ((p_module = routing_engine_find(&osm_routing_engine_list, name)) != NULL) {
		return (f_re_setup_t)p_module->setup;
	}

	return NULL;
}

static void append_routing_engine(osm_subn_t *subn, struct osm_routing_engine *routing_engine)
{
	struct osm_routing_engine *r;

	routing_engine->next = NULL;

	if (!subn->routing_engine_list) {
		subn->routing_engine_list = routing_engine;
		return;
	}

	r = subn->routing_engine_list;
	while (r->next) {
		r = r->next;
	}

	r->next = routing_engine;
}

static const char *routing_engine_get_name(struct osm_routing_engine *re)
{
	return re->name;
}

static struct osm_routing_engine_attributes routing_engine_get_attributes(struct osm_routing_engine *re)
{
	return re->attributes;
}

static void update_routing_engine(struct osm_routing_engine *cur, struct osm_routing_engine *last)
{
	struct osm_routing_engine *next = cur->next;

	if (!last)
		return; /* no last routing engine */

	/*
	 * Since function overrides 'cur', we need to make sure 'cur' context is
	 * released.
	 *
	 * We also need to free the old alloced name and strdup the new name. (See setup_routing_engine.)
	 */
	if (cur->context && cur->destroy)
		cur->destroy(cur->context);

	if (cur->name) {
		free(cur->name);
		cur->name = NULL;
	}

	memcpy(cur, last, sizeof(*cur));
	cur->name = strdup(last->name);

	/* restore next */
	cur->next = next;
}

/* -------------------------------------------------------------------------------------------------
 *
 * Routing engine modules.
 */

extern int osm_ucast_minhop_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_updn_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_dnup_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_file_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_ftree_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_pqft_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_lash_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_dor_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_torus2QoS_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_sssp_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_dfsssp_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_chain_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_dfp_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_kdor_hc_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_kdor_ghc_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_dfp2_setup(struct osm_routing_engine *, osm_opensm_t *);

extern int osm_ucast_armgr_lag_setup(osm_adv_routing_engine_t *, osm_subn_t *);
extern int osm_ucast_armgr_tree_setup(osm_adv_routing_engine_t *, osm_subn_t *);
extern int osm_ucast_armgr_minhop_setup(osm_adv_routing_engine_t *, osm_subn_t *);

extern int osm_ucast_ar_updn_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_ar_ftree_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_ar_torus_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_ar_dor_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_ar_minhop_setup(struct osm_routing_engine *, osm_opensm_t *);
extern int osm_ucast_auto_engine_setup(struct osm_routing_engine *, osm_opensm_t *);

/*
 * Routing engine modules.
 */
static re_module_t routing_modules[] = {
        {  .name = "minhop",     .setup = osm_ucast_minhop_setup,       .type = OSM_ROUTING_ENGINE_TYPE_MINHOP },
        {  .name = "updn",       .setup = osm_ucast_updn_setup,         .type = OSM_ROUTING_ENGINE_TYPE_UPDN },
        {  .name = "dnup",       .setup = osm_ucast_dnup_setup,         .type = OSM_ROUTING_ENGINE_TYPE_DNUP },
        {  .name = "file",       .setup = osm_ucast_file_setup,         .type = OSM_ROUTING_ENGINE_TYPE_FILE },
        {  .name = "ftree",      .setup = osm_ucast_ftree_setup,        .type = OSM_ROUTING_ENGINE_TYPE_FTREE },
        {  .name = "pqft",       .setup = osm_ucast_pqft_setup,         .type = OSM_ROUTING_ENGINE_TYPE_PQFT },
        {  .name = "lash",       .setup = osm_ucast_lash_setup,         .type = OSM_ROUTING_ENGINE_TYPE_LASH },
        {  .name = "dor",        .setup = osm_ucast_dor_setup,          .type = OSM_ROUTING_ENGINE_TYPE_DOR },
        {  .name = "torus-2QoS", .setup = osm_ucast_torus2QoS_setup,    .type = OSM_ROUTING_ENGINE_TYPE_TORUS_2QOS },
        {  .name = "dfsssp",     .setup = osm_ucast_dfsssp_setup,       .type = OSM_ROUTING_ENGINE_TYPE_DFSSSP },
        {  .name = "sssp",       .setup = osm_ucast_sssp_setup,         .type = OSM_ROUTING_ENGINE_TYPE_SSSP },
        {  .name = "chain",      .setup = osm_ucast_chain_setup,        .type = OSM_ROUTING_ENGINE_TYPE_CHAIN },
        {  .name = "dfp",        .setup = osm_ucast_dfp_setup,          .type = OSM_ROUTING_ENGINE_TYPE_DFP },
        {  .name = "dfp2",       .setup = osm_ucast_dfp2_setup,         .type = OSM_ROUTING_ENGINE_TYPE_DFP2 },
        {  .name = "ar_dor",     .setup = osm_ucast_ar_dor_setup,       .type = OSM_ROUTING_ENGINE_TYPE_AR_DOR },
        {  .name = "ar_updn",    .setup = osm_ucast_ar_updn_setup,      .type = OSM_ROUTING_ENGINE_TYPE_AR_UPDN },
        {  .name = "ar_ftree",   .setup = osm_ucast_ar_ftree_setup,     .type = OSM_ROUTING_ENGINE_TYPE_AR_FTREE },
        {  .name = "ar_torus",   .setup = osm_ucast_ar_torus_setup,     .type = OSM_ROUTING_ENGINE_TYPE_AR_TORUS },
        {  .name = "kdor-hc",    .setup = osm_ucast_kdor_hc_setup,      .type = OSM_ROUTING_ENGINE_TYPE_KDOR_HC },
        {  .name = "kdor-ghc",   .setup = osm_ucast_kdor_ghc_setup,     .type = OSM_ROUTING_ENGINE_TYPE_KDOR_GHC },
        {  .name = "ar_minhop",  .setup = osm_ucast_ar_minhop_setup,    .type = OSM_ROUTING_ENGINE_TYPE_AR_MINHOP },
        {  .name = "auto",       .setup = osm_ucast_auto_engine_setup,  .type = OSM_ROUTING_ENGINE_TYPE_AUTO },
        {  .name = NULL,         .setup = NULL,                         .type = OSM_ROUTING_ENGINE_TYPE_NONE },
};

/* -------------------------------------------------------------------------------------------------
 *
 * Routing engine API functions.
 */

void osm_routing_engine_init(void)
{
	re_module_t     *p_module;

	cl_qlist_init(&osm_routing_engine_list);

	for (p_module = routing_modules; p_module->name != NULL; p_module++) {
		cl_qlist_insert_tail(&osm_routing_engine_list, (cl_list_item_t *)p_module);
	}
}

osm_routing_engine_type_t osm_routing_engine_type(struct osm_routing_engine *p_re)
{
	return (p_re != NULL) ? p_re->type : OSM_ROUTING_ENGINE_TYPE_NONE;
}

const char *osm_routing_engine_type_str(osm_routing_engine_type_t type)
{
	re_module_t     		*p_module;

	foreach_routing_engine(&osm_routing_engine_list, p_module) {
		if (p_module->type == type) {
			return p_module->name;
		}
	}

	return "none";
}

const char *osm_routing_engine_name(struct osm_routing_engine *p_re)
{
	return (p_re != NULL) ? p_re->name : "none";
}

boolean_t osm_routing_engine_cmp(struct osm_routing_engine *p_re1, struct osm_routing_engine *p_re2)
{
	if ((p_re1 != NULL) && (p_re2 != NULL)) {
		return (strcmp(p_re1->name, p_re2->name) == 0) ? TRUE : FALSE;
	} else if ((p_re1 == NULL) && (p_re2 == NULL)) {
		return TRUE;
	} else {
		return FALSE;
	}
}

int osm_routing_engine_create(IN osm_opensm_t *p_osm, IN const char *name, f_re_setup_t setup_function)
{
	re_module_t     		*p_module;
	osm_routing_engine_type_t	max_type;

	/*
	 * Do we already have a module by this name?
	 */
	if ((p_module = routing_engine_find(&osm_routing_engine_list, name)) != NULL) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR A904: routing engine named \'%s\' already exists\n", name);
		return -1;
	}

	/*
	 * Get a type for this routing engine.
	 */
	max_type = LOADABLE_ROUTING_ENGINE_TYPE_BASE-1;
	foreach_routing_engine(&osm_routing_engine_list, p_module) {
		if (p_module->type > max_type) {
			max_type = p_module->type;
		}
	}

	/*
	 * Allocate the routing engine module and initialize it.
	 */
	if ((p_module = calloc(1, sizeof(re_module_t))) == NULL) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR A905: memory allocation failed\n");
		return -1;
	}

	p_module->name = strdup(name);
	p_module->setup = setup_function;
	p_module->type = max_type+1;
	cl_qlist_insert_tail(&osm_routing_engine_list, (cl_list_item_t *)p_module);

	OSM_LOG(&p_osm->log, OSM_LOG_INFO, "routing engine \'%s\' created with type %d\n", name, p_module->type);
	return 0;
}

void osm_routing_engine_free(struct osm_routing_engine *p_re)
{
	if (p_re->name) {
		free(p_re->name);
		p_re->name = NULL;
	}
	free(p_re);
}

struct osm_routing_engine *setup_routing_engine(osm_subn_t *subn, const char *name)
{
	re_module_t    			*p_module;
	struct osm_routing_engine	*re;
	f_re_setup_t			setup_function;

	if (!strcmp(name, "no_fallback")) {
		subn->no_fallback_routing_engine = TRUE;
		return NULL;
	}

	if ((p_module = routing_engine_find(&osm_routing_engine_list, name)) == NULL) {
		OSM_LOG(&subn->p_osm->log, OSM_LOG_ERROR, "ERR A906: no routing engine named \'%s\' found\n", name);
		return NULL;
	}

	/*
	 * Verify that the routing engine name is known.
	 */
	if ((setup_function = osm_routing_engine_setup(name)) == NULL) {
		OSM_LOG(&subn->p_osm->log, OSM_LOG_ERROR, "ERR A901: cannot find or setup routing engine \'%s\'\n", name);
		return NULL;
	}

	/*
	 * Allocate the routing engine.
	 */
	if ((re = calloc(1, sizeof(struct osm_routing_engine))) == NULL) {
		OSM_LOG(&subn->p_osm->log, OSM_LOG_ERROR, "ERR A902: memory allocation failed\n");
		return NULL;
	}

	re->get_name = routing_engine_get_name;
	re->get_attributes = routing_engine_get_attributes;
	re->name = strdup(name);
	re->type = p_module->type;

	/*
	 * Setup the routing engine.
	 */
	if (setup_function(re, subn->p_osm) != 0) {
		OSM_LOG(&subn->p_osm->log, OSM_LOG_ERROR, "ERR A903: setup of routing engine \'%s\' failed\n", name);
		osm_routing_engine_free(re);
		return NULL;
	}

	if (!subn->is_auto_engine_used && !strcmp(re->name, "auto")) {
		subn->is_auto_engine_used = TRUE;
	}

	OSM_LOG(&subn->p_osm->log, OSM_LOG_DEBUG, "\'%s\' routing engine set up\n", re->name);
	return re;
}

void osm_setup_routing_engines(osm_subn_t *subn, const char *engine_names)
{
	char *name, *str, *p;
	struct osm_routing_engine *re;

	subn->is_auto_engine_used = FALSE;

	if (engine_names && *engine_names) {
		str = strdup(engine_names);
		name = strtok_r(str, ", \t\n", &p);
		while (name && *name) {
			re = setup_routing_engine(subn, name);
			if (re) {
				append_routing_engine(subn, re);
				/* update is_auto_engine_used indicator */
				if (osm_routing_engine_is_auto(re))
					subn->is_auto_engine_used = TRUE;
			}
			name = strtok_r(NULL, ", \t\n", &p);
		}
		free(str);
	}
	if (!subn->default_routing_engine) {
		subn->default_routing_engine = setup_routing_engine(subn, "minhop");
	}
}

void osm_destroy_default_routing_engine(osm_subn_t *subn)
{
	struct osm_routing_engine *r;

	r = subn->default_routing_engine;
	if (r) {
		if (r->destroy && r->context)
			r->destroy(r->context);
		osm_routing_engine_free(r);
		subn->default_routing_engine = NULL;
	}
}

void osm_destroy_routing_engines(osm_subn_t *subn)
{
	struct osm_routing_engine *r, *next;

	next = subn->routing_engine_list;
	while (next) {
		r = next;
		next = r->next;
		if (r != subn->default_routing_engine) {
			if (r->destroy)
				r->destroy(r->context);
			osm_routing_engine_free(r);
		} else /*do not free default_routing_engine */
			r->next = NULL;
	}
	subn->routing_engine_list = NULL;
}

void osm_update_routing_engines(osm_subn_t *subn, const char *engine_names)
{
	struct osm_routing_engine *r, *l;

	/* find used routing engine and save as last */
	l = r = subn->routing_engine_list;

	if (!r)
		goto config_engines;

	/* skip if routing_engine_used not set */
	if (subn->routing_engine_used == NULL)
		goto destroy_engines;

	if (osm_routing_engine_cmp(subn->routing_engine_used, r)) {
		subn->last_routing_engine = r;
		subn->routing_engine_list = r->next;
	} else while ((r = r->next)) {
		if (osm_routing_engine_cmp(subn->routing_engine_used, r)) {
			subn->last_routing_engine = r;
			l->next = r->next;
			break;
		}
		l = r;
	}

destroy_engines:
	/* cleanup prev routing engine list and replace with current list */
	osm_destroy_routing_engines(subn);

config_engines:
	osm_setup_routing_engines(subn, engine_names);

	if (subn->routing_engine_used == NULL)
		return;

	if (subn->routing_engine_used == subn->default_routing_engine)
		return;

	/* check if last routing engine exist in new list and update callbacks */
	r = subn->routing_engine_list;
	while (r) {
		if (osm_routing_engine_cmp(subn->routing_engine_used, r)) {
			update_routing_engine(r, subn->last_routing_engine);
			subn->last_routing_engine = NULL;
			osm_routing_engine_free(subn->routing_engine_used);
			subn->routing_engine_used = r;
			break;
		}
		r = r->next;
	}
}

/* -------------------------------------------------------------------------------------------------
 *
 * Advanced routing engine private definitions.
 */

static cl_qlist_t       osm_adv_routing_engine_list;

typedef int (*f_adv_re_setup_t)(osm_adv_routing_engine_t *, osm_subn_t *);

extern int osm_ucast_armgr_lag_setup(osm_adv_routing_engine_t *, osm_subn_t *);
extern int osm_ucast_armgr_tree_setup(osm_adv_routing_engine_t *, osm_subn_t *);
extern int osm_ucast_armgr_minhop_setup(osm_adv_routing_engine_t *, osm_subn_t *);

/*
 * Advanced routing engine modules.
 */
struct adv_routing_engine_module {
        char *name;
        f_adv_re_setup_t setup;
};

static re_module_t adv_routing_modules[] = {
	{  .name = "none",       .setup = NULL                          },
	{  .name = "ar_lag",     .setup = osm_ucast_armgr_lag_setup     },
	{  .name = "ar_tree",    .setup = osm_ucast_armgr_tree_setup    },
	{  .name = "ar_minhop",  .setup = osm_ucast_armgr_minhop_setup  },
	{  .name = NULL,         .setup = NULL                          },
};

static f_adv_re_setup_t osm_adv_routing_engine_setup(IN const char *name)
{
	re_module_t             *p_module;

	if ((p_module = routing_engine_find(&osm_adv_routing_engine_list, name)) != NULL) {
		return (f_adv_re_setup_t)p_module->setup;
	}

	return NULL;
}

static void adv_routing_engine_free(osm_adv_routing_engine_t *p_adv_re)
{
	if (p_adv_re->name != NULL) {
		free(p_adv_re->name);
		p_adv_re->name = NULL;
	}
	free(p_adv_re);
}

/* -------------------------------------------------------------------------------------------------
 *
 * Advanced routing engine API.
 */

void osm_adv_routing_engine_init(void)
{
	re_module_t     *p_module;

	cl_qlist_init(&osm_adv_routing_engine_list);

	for (p_module = adv_routing_modules; p_module->name != NULL; p_module++) {
		cl_qlist_insert_tail(&osm_adv_routing_engine_list, (cl_list_item_t *)p_module);
	}
}

void osm_destroy_advanced_routing_engine(osm_subn_t *p_subn)
{
	if (p_subn->adv_routing_engine) {
		p_subn->adv_routing_engine->destroy(p_subn->adv_routing_engine->context);
		adv_routing_engine_free(p_subn->adv_routing_engine);
		p_subn->adv_routing_engine = NULL;
	}
}

void osm_create_advanced_routing_engine(osm_subn_t *p_subn, const char *name)
{
	osm_adv_routing_engine_t *p_adv_re = NULL;
	f_adv_re_setup_t setup_function;

	/*
	 * Get the setup function.
	 */
	setup_function = osm_adv_routing_engine_setup(name);
	if (setup_function == NULL) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7535: cannot find advanced routing engine \'%s\'\n", name);
		return;
	}

	/*
	 * Construct the new advanced routing engine.
	 */
	if ((p_adv_re = (osm_adv_routing_engine_t *)calloc(1, sizeof(*p_adv_re))) == NULL) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 7533: memory allocation failed\n");
		return;
	}

	/*
	 * Call the setup function.
	 */
	p_adv_re->name = strdup(name);
	if (setup_function(p_adv_re, p_subn)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7534: failed to initialize advanced routing engine module \'%s\'\n", name);
		adv_routing_engine_free(p_adv_re);
		return;
	}

	p_subn->adv_routing_engine = p_adv_re;
}

void osm_setup_advanced_routing_engine(osm_subn_t *p_subn, const char *engine_name)
{
	/* destroy current advanced routing engine module */
	osm_destroy_advanced_routing_engine(p_subn);

	if (!strcasecmp(engine_name, "none") || !strcasecmp(engine_name, "auto")) {
		return;
	}

	osm_create_advanced_routing_engine(p_subn, engine_name);
}

