/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2006-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2011 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2013 Oracle and/or its affiliates. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 * 	Implementation of osm_mpr_rcv_t.
 *	This object represents the MultiPath Record Receiver object.
 *	This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#if defined (VENDOR_RMPP_SUPPORT) && defined (DUAL_SIDED_RMPP)

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <complib/cl_qlist.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_SA_MULTIPATH_RECORD_C
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_port.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_partition.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_qos_policy.h>
#include <opensm/osm_sa.h>

#define OSM_SA_MPR_MAX_NUM_PATH        127
#define MAX_HOPS 64

#undef min
#define min(x,y)	(((x) < (y)) ? (x) : (y))

static osm_sa_item_t * mpr_rcv_get_apm_port_pair_paths(
    IN osm_sa_t * sa,
    IN const ib_multipath_rec_t * p_mpr,
    IN const osm_alias_guid_t * p_req_alias_guid,
    IN const osm_alias_guid_t * p_src_alias_guid,
    IN const osm_alias_guid_t * p_dest_alias_guid,
    IN const ib_gid_t * p_sgid,
    IN const ib_gid_t * p_dgid,
    IN int base_offs,
    IN const ib_net64_t comp_mask,
    IN cl_qlist_t * p_list)
{
	osm_sa_item_t *p_mpr_item = 0;
	uint16_t src_lid_min_ho;
	uint16_t src_lid_max_ho;
	uint16_t dest_lid_min_ho;
	uint16_t dest_lid_max_ho;
	uint16_t src_lid_ho;
	uint16_t dest_lid_ho;
	unsigned iterations;
	int src_lids, dest_lids;
	const osm_pkey_tbl_t *pkey_tbl_src, *pkey_tbl_dest, *pkey_tbl_req;

	OSM_LOG_ENTER(sa->p_log);

	OSM_LOG(sa->p_log, OSM_LOG_DEBUG, "Src port 0x%016" PRIx64 ", "
		"Dst port 0x%016" PRIx64 ", base offs %d\n",
		cl_ntoh64(p_src_alias_guid->alias_guid),
		cl_ntoh64(p_dest_alias_guid->alias_guid),
		base_offs);

	/*
	 * Check that the req_port, src_port and dest_port all share a pkey.
	 * If one of the pairs doesn't share a pkey, the path is disqualified.
	 *
	 * The check is based on pkey table associated to an alias guid since
	 * alias guid objects relate to virtual ports specific pkey tables.
	 * That way we can support path requests for virtual ports.
	 */
	pkey_tbl_src = osm_alias_guid_get_pkey_tbl(p_src_alias_guid);
	pkey_tbl_dest = osm_alias_guid_get_pkey_tbl(p_dest_alias_guid);
	pkey_tbl_req = osm_alias_guid_get_pkey_tbl(p_req_alias_guid);
	if (osm_pkey_tbl_share_pkey(sa->p_log, pkey_tbl_req, pkey_tbl_src,
				    sa->p_subn->opt.allow_both_pkeys) == FALSE) {
		OSM_LOG(sa->p_log, OSM_LOG_VERBOSE, "ERR 4522: "
			"Source and requester does not share a pkey\n");
		goto Exit;
	} else if (osm_pkey_tbl_share_pkey(sa->p_log, pkey_tbl_req, pkey_tbl_dest,
				sa->p_subn->opt.allow_both_pkeys) == FALSE) {

		OSM_LOG(sa->p_log, OSM_LOG_VERBOSE, "ERR 4523: "
			"Destination and requester does not share a pkey\n");
		goto Exit;
	} else if (osm_pkey_tbl_share_pkey(sa->p_log, pkey_tbl_src, pkey_tbl_dest,
				       sa->p_subn->opt.allow_both_pkeys) == FALSE) {
		OSM_LOG(sa->p_log, OSM_LOG_VERBOSE, "ERR 4524: "
			"Source and destination does not share a pkey\n");
		goto Exit;
	}

	osm_port_get_lid_range_ho(p_src_alias_guid->p_base_port,
				  &src_lid_min_ho, &src_lid_max_ho);
	osm_port_get_lid_range_ho(p_dest_alias_guid->p_base_port,
				  &dest_lid_min_ho, &dest_lid_max_ho);

	src_lid_ho = src_lid_min_ho;
	dest_lid_ho = dest_lid_min_ho;

	src_lids = src_lid_max_ho - src_lid_min_ho + 1;
	dest_lids = dest_lid_max_ho - dest_lid_min_ho + 1;

	src_lid_ho += base_offs % src_lids;
	dest_lid_ho += base_offs % dest_lids;

	OSM_LOG(sa->p_log, OSM_LOG_DEBUG,
		"Src LIDs [%u-%u] hashed %u, "
		"Dest LIDs [%u-%u] hashed %u\n",
		src_lid_min_ho, src_lid_max_ho, src_lid_ho,
		dest_lid_min_ho, dest_lid_max_ho, dest_lid_ho);

	iterations = min(src_lids, dest_lids);

	while (iterations--) {
		/*
		   These paths are "fully redundant"
		 */
		p_mpr_item = osm_pr_rcv_get_lid_pair_path(sa, p_mpr,
							  p_src_alias_guid,
							  p_dest_alias_guid,
							  p_sgid, NULL, p_dgid,
							  src_lid_ho, dest_lid_ho,
							  comp_mask, 0);

		if (p_mpr_item) {
			OSM_LOG(sa->p_log, OSM_LOG_DEBUG,
				"Found matching path from Src LID %u to Dest LID %u with %d hops\n",
				src_lid_ho, dest_lid_ho, p_mpr_item->resp.mpr_rec.hops);
			break;
		}

		if (++src_lid_ho > src_lid_max_ho)
			src_lid_ho = src_lid_min_ho;

		if (++dest_lid_ho > dest_lid_max_ho)
			dest_lid_ho = dest_lid_min_ho;
	}

Exit:
	OSM_LOG_EXIT(sa->p_log);
	return p_mpr_item;
}

static ib_net16_t mpr_rcv_get_gids(IN osm_sa_t * sa, IN const ib_gid_t * gids,
				   IN int ngids, IN int is_sgid,
				   OUT osm_alias_guid_t ** pp_alias_guid,
				   OUT const ib_gid_t ** pp_remote_dgids,
				   OUT boolean_t * router_path)
{
	osm_alias_guid_t *p_alias_guid;
	ib_net16_t ib_status = IB_SUCCESS;
	int i;

	OSM_LOG_ENTER(sa->p_log);

	for (i = 0; i < ngids; i++, gids++) {
		pp_alias_guid[i] = NULL;

		if (!ib_gid_is_link_local(gids)) {
			if (!is_sgid &&
			    !is_equal_subnet_prefix(ib_gid_get_subnet_prefix(gids),
						    sa->p_subn->opt.subnet_prefix)) {
				if (!ib_gid_is_site_local(gids)) {
					/*
					 * it's global scope. Not supported yet
					 * This 'error' is the client's fault (bad gid)
					 * so don't enter it as an error in our own log.
					 * Return an error response to the client.
					 */
					   OSM_LOG(sa->p_log, OSM_LOG_VERBOSE,
						   "ERR 4521: Global DGID subnet prefix"
						   " 0x%016" PRIx64 " is not supported\n",
						   cl_ntoh64(gids->unicast.prefix));
					ib_status = IB_SA_MAD_STATUS_INVALID_GID;
					goto Exit;
				}

				if (pp_remote_dgids) {
					*pp_remote_dgids = gids;
					*router_path = TRUE;
				}

			} else if ((is_sgid && ib_gid_is_multicast(gids)) ||
			    (!is_equal_subnet_prefix(ib_gid_get_subnet_prefix(gids),
						     sa->p_subn->opt.subnet_prefix))) {
				/*
				   This 'error' is the client's fault (bad gid)
				   so don't enter it as an error in our own log.
				   Return an error response to the client.
				 */
				OSM_LOG(sa->p_log, OSM_LOG_VERBOSE, "ERR 451B: "
					"SGID 0x%016" PRIx64
					" is multicast or non local subnet prefix\n",
					cl_ntoh64(gids->unicast.prefix));

				ib_status = IB_SA_MAD_STATUS_INVALID_GID;
				goto Exit;
			}
		}

		/* Check GUID as part of local subnet if is SGID or local DGID */
		if (is_sgid ||
		    (ib_gid_is_link_local(gids) ||
                     is_equal_subnet_prefix(ib_gid_get_subnet_prefix(gids),
					    sa->p_subn->opt.subnet_prefix))) {
			p_alias_guid =
			    osm_get_alias_guid_by_guid(sa->p_subn,
						       gids->unicast.interface_id);
			if (!p_alias_guid) {
				/*
				   This 'error' is the client's fault (bad gid) so
				   don't enter it as an error in our own log.
				   Return an error response to the client.
				 */
				OSM_LOG(sa->p_log, OSM_LOG_ERROR, "ERR 4506: "
					"No port with GUID 0x%016" PRIx64 "\n",
					cl_ntoh64(gids->unicast.interface_id));

				ib_status = IB_SA_MAD_STATUS_INVALID_GID;
				goto Exit;
			}

			pp_alias_guid[i] = p_alias_guid;
		}

		if (pp_remote_dgids)
			pp_remote_dgids++;
	}

Exit:
	OSM_LOG_EXIT(sa->p_log);

	return ib_status;
}

static ib_net16_t mpr_rcv_get_end_points(IN osm_sa_t * sa,
					 IN const osm_madw_t * p_madw,
					 OUT osm_alias_guid_t ** pp_alias_guids,
					 OUT int *nsrc, OUT int *ndest,
					 OUT const ib_gid_t ** pp_remote_dgids,
					 OUT boolean_t * router_path)
{
	const ib_multipath_rec_t *p_mpr;
	const ib_sa_mad_t *p_sa_mad;
	ib_net64_t comp_mask;
	ib_net16_t sa_status = IB_SA_MAD_STATUS_SUCCESS;
	ib_gid_t *gids;

	OSM_LOG_ENTER(sa->p_log);

	/*
	   Determine what fields are valid and then get a pointer
	   to the source and destination port objects, if possible.
	 */
	p_sa_mad = osm_madw_get_sa_mad_ptr(p_madw);
	p_mpr = (ib_multipath_rec_t *) ib_sa_mad_get_payload_ptr(p_sa_mad);
	gids = (ib_gid_t *) p_mpr->gids;

	comp_mask = p_sa_mad->comp_mask;

	/*
	   Check a few easy disqualifying cases up front before getting
	   into the endpoints.
	 */
	*nsrc = *ndest = 0;

	if (comp_mask & IB_MPR_COMPMASK_SGIDCOUNT) {
		*nsrc = p_mpr->sgid_count;
		if (*nsrc > IB_MULTIPATH_MAX_GIDS) {
			OSM_LOG(sa->p_log, OSM_LOG_VERBOSE,
				"Number of SDIGs is larger than supported"
				" (%d > %d). Uses first %d source gids\n",
				*nsrc, IB_MULTIPATH_MAX_GIDS, IB_MULTIPATH_MAX_GIDS);

			*nsrc = IB_MULTIPATH_MAX_GIDS;
		}

		sa_status = mpr_rcv_get_gids(sa, gids, *nsrc, 1, pp_alias_guids,
					     NULL, router_path);
		if (sa_status != IB_SUCCESS)
			goto Exit;
	}

	if (comp_mask & IB_MPR_COMPMASK_DGIDCOUNT) {
		*ndest = p_mpr->dgid_count;
		if (*ndest + *nsrc > IB_MULTIPATH_MAX_GIDS) {
			OSM_LOG(sa->p_log, OSM_LOG_VERBOSE,
				"Number of GIDs is larger than supported"
				" (%d > %d). Uses first %d DGIDs\n",
				*nsrc + *ndest, IB_MULTIPATH_MAX_GIDS,
				IB_MULTIPATH_MAX_GIDS - *nsrc);

			*ndest = IB_MULTIPATH_MAX_GIDS - *nsrc;
		}
		sa_status =
		    mpr_rcv_get_gids(sa, gids + *nsrc, *ndest, 0,
				     pp_alias_guids + *nsrc,
				     &pp_remote_dgids[*nsrc], router_path);
	}

Exit:
	OSM_LOG_EXIT(sa->p_log);
	return sa_status;
}

#define hash_lids(a, b, lmc)	\
	(((((a) >> (lmc)) << 4) | ((b) >> (lmc))) % 103)

static void mpr_rcv_get_apm_paths(IN osm_sa_t * sa,
				  IN const ib_multipath_rec_t * p_mpr,
				  IN const osm_alias_guid_t *p_req_alias_guid,
				  IN osm_alias_guid_t ** _pp_alias_guids,
				  IN const ib_net64_t comp_mask,
				  IN cl_qlist_t * p_list)
{
	osm_alias_guid_t *pp_alias_guids[4];
	IN const ib_gid_t *pp_gids[4];
	osm_sa_item_t *matrix[2][2];
	int base_offs, src_lid_ho, dest_lid_ho;
	int sumA, sumB, minA, minB;
	uint8_t src_lmc, dest_lmc;
	OSM_LOG_ENTER(sa->p_log);

	/*
	 * We want to:
	 *    1. use different lid offsets (from base) for the resultant paths
	 *    to increase the probability of redundant paths or in case
	 *    of Clos - to ensure it (different offset => different spine!)
	 *    2. keep consistent paths no matter of direction and order of ports
	 *    3. distribute the lid offsets to balance the load
	 * So, we sort the ports (within the srcs, and within the dests),
	 * hash the lids of S0, D0 (after the sort), and call mpr_rcv_get_apm_port_pair_paths
	 * with base_lid for S0, D0 and base_lid + 1 for S1, D1. This way we will get
	 * always the same offsets - order independent, and make sure different spines are used.
	 * Note that the diagonals on a Clos have the same number of hops, so it doesn't
	 * really matter which diagonal we use.
	 */
	if (_pp_alias_guids[0]->p_base_port->guid <
	    _pp_alias_guids[1]->p_base_port->guid) {
		pp_alias_guids[0] = _pp_alias_guids[0];
		pp_alias_guids[1] = _pp_alias_guids[1];
		pp_gids[0] = &p_mpr->gids[0];
		pp_gids[1] = &p_mpr->gids[1];
	} else {
		pp_alias_guids[0] = _pp_alias_guids[1];
		pp_alias_guids[1] = _pp_alias_guids[0];
		pp_gids[0] = &p_mpr->gids[1];
		pp_gids[1] = &p_mpr->gids[0];
	}
	if (_pp_alias_guids[2]->p_base_port->guid <
	    _pp_alias_guids[3]->p_base_port->guid) {
		pp_alias_guids[2] = _pp_alias_guids[2];
		pp_alias_guids[3] = _pp_alias_guids[3];
		pp_gids[2] = &p_mpr->gids[2];
		pp_gids[3] = &p_mpr->gids[3];
	} else {
		pp_alias_guids[2] = _pp_alias_guids[3];
		pp_alias_guids[3] = _pp_alias_guids[2];
		pp_gids[2] = &p_mpr->gids[3];
		pp_gids[3] = &p_mpr->gids[2];
	}

	src_lid_ho = osm_port_get_base_lid(pp_alias_guids[0]->p_base_port);
	dest_lid_ho = osm_port_get_base_lid(pp_alias_guids[2]->p_base_port);
	src_lmc = osm_port_get_lmc(pp_alias_guids[0]->p_base_port);
	dest_lmc = osm_port_get_lmc(pp_alias_guids[2]->p_base_port);
	base_offs = src_lid_ho < dest_lid_ho ?
	    hash_lids(src_lid_ho, dest_lid_ho, src_lmc) :
	    hash_lids(dest_lid_ho, src_lid_ho, dest_lmc);

	matrix[0][0] =
	    mpr_rcv_get_apm_port_pair_paths(sa, p_mpr, p_req_alias_guid,
					    pp_alias_guids[0],
					    pp_alias_guids[2],
					    pp_gids[0], pp_gids[2],
					    base_offs,
					    comp_mask, p_list);
	matrix[0][1] =
	    mpr_rcv_get_apm_port_pair_paths(sa, p_mpr, p_req_alias_guid,
					    pp_alias_guids[0],
					    pp_alias_guids[3],
					    pp_gids[0], pp_gids[3],
					    base_offs,
					    comp_mask, p_list);
	matrix[1][0] =
	    mpr_rcv_get_apm_port_pair_paths(sa, p_mpr, p_req_alias_guid,
					    pp_alias_guids[1],
					    pp_alias_guids[2],
					    pp_gids[1], pp_gids[2],
					    base_offs + 1,
					    comp_mask, p_list);
	matrix[1][1] =
	    mpr_rcv_get_apm_port_pair_paths(sa, p_mpr, p_req_alias_guid,
					    pp_alias_guids[1],
					    pp_alias_guids[3],
					    pp_gids[1], pp_gids[3],
					    base_offs + 1,
					    comp_mask, p_list);

	OSM_LOG(sa->p_log, OSM_LOG_DEBUG, "APM matrix:\n"
		"\t{0,0} 0x%X->0x%X (%d)\t| {0,1} 0x%X->0x%X (%d)\n"
		"\t{1,0} 0x%X->0x%X (%d)\t| {1,1} 0x%X->0x%X (%d)\n",
		matrix[0][0] ? cl_ntoh16(matrix[0][0]->resp.mpr_rec.path_rec.slid) : 0,
		matrix[0][0] ? cl_ntoh16(matrix[0][0]->resp.mpr_rec.path_rec.dlid) : 0,
		matrix[0][0] ? matrix[0][0]->resp.mpr_rec.hops : 0,
		matrix[0][1] ? cl_ntoh16(matrix[0][1]->resp.mpr_rec.path_rec.slid) : 0,
		matrix[0][1] ? cl_ntoh16(matrix[0][1]->resp.mpr_rec.path_rec.dlid) : 0,
		matrix[0][1] ? matrix[0][1]->resp.mpr_rec.hops : 0,
		matrix[1][0] ? cl_ntoh16(matrix[1][0]->resp.mpr_rec.path_rec.slid) : 0,
		matrix[1][0] ? cl_ntoh16(matrix[1][0]->resp.mpr_rec.path_rec.dlid) : 0,
		matrix[1][0] ? matrix[1][0]->resp.mpr_rec.hops : 0,
		matrix[1][1] ? cl_ntoh16(matrix[1][1]->resp.mpr_rec.path_rec.slid) : 0,
		matrix[1][1] ? cl_ntoh16(matrix[1][1]->resp.mpr_rec.path_rec.dlid) : 0,
		matrix[1][1] ? matrix[1][1]->resp.mpr_rec.hops : 0);

	sumA = minA = sumB = minB = 0;

	/* check diagonal A {(0,0), (1,1)} */
	if (matrix[0][0]) {
		sumA += matrix[0][0]->resp.mpr_rec.hops;
		minA = matrix[0][0]->resp.mpr_rec.hops;
	}
	if (matrix[1][1]) {
		sumA += matrix[1][1]->resp.mpr_rec.hops;
		if (minA)
			minA = min(minA, matrix[1][1]->resp.mpr_rec.hops);
		else
			minA = matrix[1][1]->resp.mpr_rec.hops;
	}

	/* check diagonal B {(0,1), (1,0)} */
	if (matrix[0][1]) {
		sumB += matrix[0][1]->resp.mpr_rec.hops;
		minB = matrix[0][1]->resp.mpr_rec.hops;
	}
	if (matrix[1][0]) {
		sumB += matrix[1][0]->resp.mpr_rec.hops;
		if (minB)
			minB = min(minB, matrix[1][0]->resp.mpr_rec.hops);
		else
			minB = matrix[1][0]->resp.mpr_rec.hops;
	}

	/* and the winner is... */
	if (minA <= minB || (minA == minB && sumA < sumB)) {
		/* Diag A */
		OSM_LOG(sa->p_log, OSM_LOG_DEBUG,
			"Diag {0,0} & {1,1} is the best:\n"
			"\t{0,0} 0x%X->0x%X (%d)\t & {1,1} 0x%X->0x%X (%d)\n",
			matrix[0][0] ? cl_ntoh16(matrix[0][0]->resp.mpr_rec.path_rec.slid) : 0,
			matrix[0][0] ? cl_ntoh16(matrix[0][0]->resp.mpr_rec.path_rec.dlid) : 0,
			matrix[0][0] ? matrix[0][0]->resp.mpr_rec.hops : 0,
			matrix[1][1] ? cl_ntoh16(matrix[1][1]->resp.mpr_rec.path_rec.slid) : 0,
			matrix[1][1] ? cl_ntoh16(matrix[1][1]->resp.mpr_rec.path_rec.dlid) : 0,
			matrix[1][1] ? matrix[1][1]->resp.mpr_rec.hops : 0);
		if (matrix[0][0])
			cl_qlist_insert_tail(p_list, &matrix[0][0]->list_item);
		if (matrix[1][1])
			cl_qlist_insert_tail(p_list, &matrix[1][1]->list_item);
		free(matrix[0][1]);
		free(matrix[1][0]);
	} else {
		/* Diag B */
		OSM_LOG(sa->p_log, OSM_LOG_DEBUG,
			"Diag {0,1} & {1,0} is the best:\n"
			"\t{0,1} 0x%X->0x%X (%d)\t & {1,0} 0x%X->0x%X (%d)\n",
			matrix[0][1] ? cl_ntoh16(matrix[0][1]->resp.mpr_rec.path_rec.slid) : 0,
			matrix[0][1] ? cl_ntoh16(matrix[0][1]->resp.mpr_rec.path_rec.dlid) : 0,
			matrix[0][1] ? matrix[0][1]->resp.mpr_rec.hops : 0,
			matrix[1][0] ? cl_ntoh16(matrix[1][0]->resp.mpr_rec.path_rec.slid) : 0,
			matrix[1][0] ? cl_ntoh16(matrix[1][0]->resp.mpr_rec.path_rec.dlid) : 0,
			matrix[1][0] ? matrix[1][0]->resp.mpr_rec.hops : 0);
		if (matrix[0][1])
			cl_qlist_insert_tail(p_list, &matrix[0][1]->list_item);
		if (matrix[1][0])
			cl_qlist_insert_tail(p_list, &matrix[1][0]->list_item);
		free(matrix[0][0]);
		free(matrix[1][1]);
	}

	OSM_LOG_EXIT(sa->p_log);
}

static void mpr_rcv_process_pairs(IN osm_sa_t * sa,
				  IN ib_sa_mad_t *p_sa_mad,
				  IN const ib_multipath_rec_t * p_mpr,
				  IN const osm_alias_guid_t *p_req_alias_guid,
				  IN osm_alias_guid_t ** pp_alias_guids,
				  IN const int nsrc, IN int ndest,
				  IN ib_net64_t comp_mask,
				  IN cl_qlist_t * p_list,
				  IN const ib_gid_t ** pp_remote_dgids)
{
	osm_alias_guid_t **pp_src_alias_guid, **pp_es;
	osm_alias_guid_t **pp_dest_alias_guid, **pp_ed;
	uint32_t max_paths, num_paths, total_paths = 0;
	const ib_gid_t *p_sgid, *p_dgid;
	uint8_t src_index, dest_index;

	OSM_LOG_ENTER(sa->p_log);

	if (comp_mask & IB_MPR_COMPMASK_NUMBPATH)
		max_paths = p_mpr->num_path & 0x7F;
	else
		max_paths = OSM_SA_MPR_MAX_NUM_PATH;

	src_index = 0;
	for (pp_src_alias_guid = pp_alias_guids, pp_es = pp_alias_guids + nsrc;
	     pp_src_alias_guid < pp_es; pp_src_alias_guid++) {
		p_sgid = &p_mpr->gids[src_index];
		dest_index = nsrc;
		for (pp_dest_alias_guid = pp_es, pp_ed = pp_es + ndest;
		     pp_dest_alias_guid < pp_ed; pp_dest_alias_guid++) {
			p_dgid = &p_mpr->gids[dest_index];
			if (pp_remote_dgids[dest_index]) {
				osm_pr_rcv_process_routers(sa, p_sa_mad, NULL,
							   p_req_alias_guid,
							   *pp_src_alias_guid,
							   p_sgid,
							   pp_remote_dgids[dest_index],
							   p_dgid, p_list);
				num_paths = cl_qlist_count(p_list) - total_paths;
			} else {
				num_paths =
				    osm_pr_rcv_get_port_pair_paths(sa, p_sa_mad,
								   p_req_alias_guid,
								   *pp_src_alias_guid,
								   *pp_dest_alias_guid,
								   p_sgid, NULL,
								   p_dgid, p_list,
								   SA_INVALID_LID_OFFSET,
								   max_paths - total_paths);
			}

			total_paths += num_paths;
			OSM_LOG(sa->p_log, OSM_LOG_DEBUG,
				"%d paths %d total paths %d max paths\n",
				num_paths, total_paths, max_paths);
			/* Just take first NumbPaths found */
			if (total_paths >= max_paths)
				goto Exit;
			dest_index++;
		}
		src_index++;
	}

Exit:
	OSM_LOG_EXIT(sa->p_log);
}

void osm_mpr_rcv_process(IN void *context, IN void *data)
{
	osm_sa_t *sa = context;
	osm_madw_t *p_madw = data;
	const ib_multipath_rec_t *p_mpr;
	ib_sa_mad_t *p_sa_mad;
	osm_alias_guid_t *pp_alias_guids[IB_MULTIPATH_MAX_GIDS];
	const ib_gid_t *pp_remote_dgids[IB_MULTIPATH_MAX_GIDS];
	cl_qlist_t pr_list;
	ib_net16_t sa_status;
	int nsrc, ndest;
	uint8_t rate, mtu;
	osm_alias_guid_t *p_req_alias_guid;
	osm_mad_addr_t *p_mad_addr;
	boolean_t router_path = FALSE;

	OSM_LOG_ENTER(sa->p_log);

	OSM_ASSERT(p_madw);
	p_sa_mad = osm_madw_get_sa_mad_ptr(p_madw);
	p_mpr = (ib_multipath_rec_t *) ib_sa_mad_get_payload_ptr(p_sa_mad);

	OSM_ASSERT(p_sa_mad->attr_id == IB_MAD_ATTR_MULTIPATH_RECORD);

	if ((p_sa_mad->rmpp_flags & IB_RMPP_FLAG_ACTIVE) != IB_RMPP_FLAG_ACTIVE) {
		OSM_LOG(sa->p_log, OSM_LOG_ERROR, "ERR 4510: "
			"Invalid request since RMPP_FLAG_ACTIVE is not set\n");
		osm_sa_send_error(sa, p_madw, IB_SA_MAD_STATUS_REQ_INVALID);
		goto Exit;
	}

	/* we only support SubnAdmGetMulti method */
	if (p_sa_mad->method != IB_MAD_METHOD_GETMULTI) {
		OSM_LOG(sa->p_log, OSM_LOG_ERROR, "ERR 4513: "
			"Unsupported Method (%s)\n",
			ib_get_sa_method_str(p_sa_mad->method));
		osm_sa_send_error(sa, p_madw, IB_MAD_STATUS_UNSUP_METHOD_ATTR);
		goto Exit;
	}

	if (OSM_LOG_IS_ACTIVE_V2(sa->p_log, OSM_LOG_DEBUG))
		osm_dump_multipath_record_v2(sa->p_log, p_mpr, FILE_ID, OSM_LOG_DEBUG);

	/* Make sure required components (S/DGIDCount) are supplied */
	if (!(p_sa_mad->comp_mask & IB_MPR_COMPMASK_SGIDCOUNT) ||
	    !(p_sa_mad->comp_mask & IB_MPR_COMPMASK_DGIDCOUNT)) {
		osm_sa_send_error(sa, p_madw, IB_SA_MAD_STATUS_INSUF_COMPS);
		goto Exit;
	}

	/* Validate rate if supplied */
	if ((p_sa_mad->comp_mask & IB_MPR_COMPMASK_RATESELEC) &&
	    (p_sa_mad->comp_mask & IB_MPR_COMPMASK_RATE)) {
		rate = ib_multipath_rec_rate(p_mpr);
		if (!ib_rate_is_valid(rate)) {
			osm_sa_send_error(sa, p_madw,
					  IB_SA_MAD_STATUS_REQ_INVALID);
			goto Exit;
		}
	}
	/* Validate MTU if supplied */
	if ((p_sa_mad->comp_mask & IB_MPR_COMPMASK_MTUSELEC) &&
	    (p_sa_mad->comp_mask & IB_MPR_COMPMASK_MTU)) {
		mtu = ib_multipath_rec_mtu(p_mpr);
		if (!ib_mtu_is_valid(mtu)) {
			osm_sa_send_error(sa, p_madw,
					  IB_SA_MAD_STATUS_REQ_INVALID);
			goto Exit;
		}
	}

	/* Make sure either none or both ServiceID parameters are supplied */
	if ((p_sa_mad->comp_mask & IB_MPR_COMPMASK_SERVICEID) != 0 &&
	    (p_sa_mad->comp_mask & IB_MPR_COMPMASK_SERVICEID) !=
	     IB_MPR_COMPMASK_SERVICEID) {
		osm_sa_send_error(sa, p_madw, IB_SA_MAD_STATUS_INSUF_COMPS);
		goto Exit;
	}

	cl_qlist_init(&pr_list);

	/*
	 * Most SA functions (including this one) are read-only on the
	 * subnet object, so we grab the lock non-exclusively.
	 * Query access routing calculation structures, hence we lock
	 * routing_lock as well.
	 */
	osm_sa_acquire_locks(sa);

	/* Find the requester port */
	p_mad_addr = osm_madw_get_mad_addr_ptr(p_madw);
	p_req_alias_guid = osm_get_alias_guid_by_mad_addr(sa->p_log, sa->p_subn,
							  p_mad_addr);
	if (!p_req_alias_guid) {
		OSM_LOG(sa->p_log, OSM_LOG_ERROR, "ERR 450A: "
			"Received MultiPathRecord request from unknown port\n");
		osm_sa_release_locks(sa);
		goto Exit;
	}

	if (!osm_sa_validate_request_rate(sa, p_madw, p_req_alias_guid)) {
		osm_sa_release_locks(sa);
		goto Exit;
	}

	OSM_LOG(sa->p_log, OSM_LOG_DEBUG,
		"Requester port GUID 0x%" PRIx64 "\n",
		cl_ntoh64(p_req_alias_guid->alias_guid));

	/*
	 * Notic: if query's DGID is not of local subnet, pp_remote_dgids holds
	 * the DGID on the relevant entry. Otherwise, entry is NULL!
	 * Thus, each entry on pp_remote_dgids indicates whether to use router
	 * flow for each path that uses entry's GID as DGID.
 	 */
	memset(pp_remote_dgids, 0, sizeof(ib_gid_t *) * IB_MULTIPATH_MAX_GIDS);
	sa_status = mpr_rcv_get_end_points(sa, p_madw, pp_alias_guids,
					   &nsrc, &ndest, pp_remote_dgids,
					   &router_path);

	if (sa_status != IB_SA_MAD_STATUS_SUCCESS || !nsrc || !ndest) {
		osm_sa_release_locks(sa);
		if (sa_status == IB_SA_MAD_STATUS_SUCCESS && (!nsrc || !ndest))
			OSM_LOG(sa->p_log, OSM_LOG_ERROR, "ERR 4512: "
				"mpr_rcv_get_end_points failed, # GIDs found; "
				"src %d; dest %d)\n", nsrc, ndest);
		if (sa_status == IB_SA_MAD_STATUS_SUCCESS)
			osm_sa_send_error(sa, p_madw,
					  IB_SA_MAD_STATUS_REQ_INVALID);
		else
			osm_sa_send_error(sa, p_madw, sa_status);
		goto Exit;
	}

	/* APM request */
	/*
	 * router_path is TRUE if any of query's DGIDs is not local.
	 * Meaning, destination is remote. Hence, avoid APM because
	 * cannot assure disjoint paths beyond the router.
	 */
	if (!router_path && (nsrc == 2 && ndest == 2 && (p_mpr->num_path & 0x7F) == 2))
		mpr_rcv_get_apm_paths(sa, p_mpr, p_req_alias_guid, pp_alias_guids,
				      p_sa_mad->comp_mask, &pr_list);
	else
		mpr_rcv_process_pairs(sa, p_sa_mad, p_mpr, p_req_alias_guid, pp_alias_guids,
				      nsrc, ndest, p_sa_mad->comp_mask,
				      &pr_list, pp_remote_dgids);

	osm_sa_release_locks(sa);

	/* o15-0.2.7: If MultiPath is supported, then SA shall respond to a
	   SubnAdmGetMulti() containing a valid MultiPathRecord attribute with
	   a set of zero or more PathRecords satisfying the constraints
	   indicated in the MultiPathRecord received. The PathRecord Attribute
	   ID shall be used in the response.
	 */
	p_sa_mad->attr_id = IB_MAD_ATTR_PATH_RECORD;
	osm_sa_respond(sa, p_madw, sizeof(ib_path_rec_t), &pr_list);

Exit:
	OSM_LOG_EXIT(sa->p_log);
}
#endif
